'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import patterns


class EventTest(test.TestCase):
    def setUp(self):
        self.event = patterns.Event('eventtype', self, 'some value')

    def testEqualWhenAllValuesAreEqual(self):
        self.assertEqual(self.event,
                         patterns.Event('eventtype', self, 'some value'))

    def testUnequalWhenValuesAreDifferent(self):
        self.assertNotEqual(self.event,
                            patterns.Event('eventtype', self, 'other value'))
    
    def testUnequalWhenTypesAreDifferent(self):
        self.assertNotEqual(self.event,
                            patterns.Event('other type', self, 'some value'))

    def testUnequalWhenSourcesAreDifferent(self):
        self.assertNotEqual(self.event,
                            patterns.Event('eventtype', None, 'some value'))
        
    def testEventWithoutType(self):
        event = patterns.Event()
        self.assertEqual(set(), event.types())
                
    def testEventWithoutSources(self):
        event = patterns.Event('eventtype')
        self.assertEqual(set(), event.sources())

    def testEventSources(self):
        self.assertEqual(set([self]), self.event.sources())

    def testEventValue(self):
        self.assertEqual('some value', self.event.value())

    def testEventValues(self):
        self.assertEqual(('some value',), self.event.values())

    def testEventValueForSpecificSource(self):
        self.assertEqual('some value', self.event.value(self))

    def testEventValuesForSpecificSource(self):
        self.assertEqual(('some value',), self.event.values(self))
        
    def testAddSource(self):
        self.event.addSource('source')
        self.assertEqual(set([self, 'source']), self.event.sources())
        
    def testAddExistingSource(self):
        self.event.addSource(self)
        self.assertEqual(set([self]), self.event.sources())
        
    def testAddSourceAndValue(self):
        self.event.addSource('source', 'value')
        self.assertEqual('value', self.event.value('source'))

    def testAddSourceAndValues(self):
        self.event.addSource('source', 'value1', 'value2')
        self.assertEqual(set(['value1', 'value2']), set(self.event.values('source')))
        
    def testExistingSourceAndValue(self):
        self.event.addSource(self, 'new value')
        self.assertEqual(set(['some value', 'new value']), set(self.event.values()))

    def testEventTypes(self):
        self.assertEqual(set(['eventtype']), self.event.types())
        
    def testAddSourceForSpecificType(self):
        self.event.addSource(self, type='another eventtype')
        self.assertEqual(set(['eventtype', 'another eventtype']), 
                         self.event.types())
        
    def testGetSourcesForSpecificType(self):
        self.assertEqual(set([self]), self.event.sources('eventtype'))
        
    def testGetSourcesForSpecificTypes(self):
        self.event.addSource('source', type='another eventtype')
        self.assertEqual(set([self, 'source']), self.event.sources(*self.event.types()))
        
    def testGetSourcesForNonExistingEventType(self):
        self.assertEqual(set(), self.event.sources('unused eventType'))
        
    def testGetAllSourcesAfterAddingSourceForSpecificType(self):
        self.event.addSource('source', type='another eventtype')
        self.assertEqual(set([self, 'source']), self.event.sources())
        
    def testAddSourceAndValueForSpecificType(self):
        self.event.addSource('source', 'value', type='another eventtype')
        self.assertEqual('value', self.event.value('source'))
        
    def testAddSourceAndValuesForSpecificType(self):
        self.event.addSource('source', 'value1', 'value2', 
                             type='another eventtype')
        self.assertEqual(set(['value1', 'value2']), set(self.event.values('source')))
        
    def testAddExistingSourceToAnotherType(self):
        self.event.addSource(self, type='another eventtype')
        self.assertEqual(set([self]), self.event.sources())

    def testAddExistingSourceWithValueToTypeDoesNotRemoveValueForEarlierType(self):
        self.event.addSource(self, 'value for another eventtype', 
                             type='another eventtype')
        self.assertEqual('some value', self.event.value(self, type='eventtype'))

    def testAddExistingSourceWithValueToType(self):
        self.event.addSource(self, 'value for another eventtype', 
                             type='another eventtype')
        self.assertEqual('value for another eventtype', 
                         self.event.value(self, type='another eventtype'))
        
    def testSubEventForOneTypeWhenEventHasOneType(self):
        self.assertEqual(self.event, self.event.subEvent((self.event.type(), self)))

    def testSubEventForOneTypeWhenEventHasTwoTypes(self):
        self.event.addSource('source', type='another eventtype')
        expectedEvent = patterns.Event('eventtype', self, 'some value')
        self.assertEqual(expectedEvent, self.event.subEvent(('eventtype', self)))

    def testSubEventForTwoTypesWhenEventHasTwoTypes(self):
        self.event.addSource('source', type='another eventtype')
        args = [('eventtype', self), ('another eventtype', 'source')]
        self.assertEqual(self.event, self.event.subEvent(*args)) # pylint: disable=W0142

    def testSubEventForTypeThatIsNotPresent(self):
        self.assertEqual(patterns.Event(), 
                         self.event.subEvent(('missing eventtype', self)))
        
    def testSubEventForOneSourceWhenEventHasOneSource(self):
        self.assertEqual(self.event, self.event.subEvent(('eventtype', self)))
        
    def testSubEventForUnspecifiedSource(self):
        self.assertEqual(self.event, self.event.subEvent(('eventtype', None)))

    def testSubEventForUnspecifiedSourceAndSpecifiedSources(self):
        self.assertEqual(self.event, self.event.subEvent(('eventtype', self), 
                                                         ['eventtype', None]))
        
    def testSubEventForSourceThatIsNotPresent(self):
        self.assertEqual(patterns.Event(), 
                         self.event.subEvent(('eventtype', 'missing source')))

    def testSubEventForSourceThatIsNotPresentForSpecifiedType(self):
        self.event.addSource('source', type='another eventtype')
        self.assertEqual(patterns.Event(), 
                         self.event.subEvent(('eventtype', 'source')))
                

class ObservableCollectionFixture(test.TestCase):
    def setUp(self):
        self.collection = self.createObservableCollection()
        patterns.Publisher().registerObserver(self.onAdd, 
            eventType=self.collection.addItemEventType(),
            eventSource=self.collection)
        patterns.Publisher().registerObserver(self.onRemove,
            eventType=self.collection.removeItemEventType(),
            eventSource=self.collection)
        self.receivedAddEvents = []
        self.receivedRemoveEvents = []
        
    def createObservableCollection(self):
        raise NotImplementedError # pragma: no cover

    def onAdd(self, event):
        self.receivedAddEvents.append(event)

    def onRemove(self, event):
        self.receivedRemoveEvents.append(event)


class ObservableCollectionTestsMixin(object):
    def testCollectionEqualsItself(self):
        self.failUnless(self.collection == self.collection)

    def testCollectionDoesNotEqualOtherCollections(self):
        self.failIf(self.collection == self.createObservableCollection())

    def testAppend(self):
        self.collection.append(1)
        self.failUnless(1 in self.collection)
        
    def testAppend_Notification(self):
        self.collection.append(1)
        self.assertEqual(1, self.receivedAddEvents[0].value())

    def testExtend(self):
        self.collection.extend([1, 2])
        self.failUnless(1 in self.collection and 2 in self.collection)
        
    def testExtend_Notification(self):
        self.collection.extend([1, 2, 3])
        self.assertEqual((1, 2, 3), self.receivedAddEvents[0].values())

    def testExtend_NoNotificationWhenNoItems(self):
        self.collection.extend([])
        self.failIf(self.receivedAddEvents)

    def testRemove(self):
        self.collection.append(1)
        self.collection.remove(1)
        self.failIf(self.collection)

    def testRemove_Notification(self):
        self.collection.append(1)
        self.collection.remove(1)
        self.assertEqual(1, self.receivedRemoveEvents[0].value())
    
    def testRemovingAnItemNotInCollection_CausesException(self):
        try:
            self.collection.remove(1)
            self.fail('Expected ValueError or KeyError') # pragma: no cover
        except (ValueError, KeyError):
            pass

    def testRemovingAnItemNotInCollection_CausesNoNotification(self):
        try:
            self.collection.remove(1)
        except (ValueError, KeyError):
            pass
        self.failIf(self.receivedRemoveEvents)

    def testRemoveItems(self):
        self.collection.extend([1, 2, 3])
        self.collection.removeItems([1, 2])
        self.failIf(1 in self.collection or 2 in self.collection)

    def testRemoveItems_Notification(self):
        self.collection.extend([1, 2, 3])
        self.collection.removeItems([1, 2])
        self.assertEqual((1, 2), self.receivedRemoveEvents[0].values())

    def testRemoveItems_NoNotificationWhenNoItems(self):
        self.collection.extend([1, 2, 3])
        self.collection.removeItems([])
        self.failIf(self.receivedRemoveEvents)
        
    def testClear(self):
        self.collection.extend([1, 2, 3])
        self.collection.clear()
        self.failIf(self.collection)
        
    def testClear_Notification(self):
        self.collection.extend([1, 2, 3])
        self.collection.clear()
        self.assertEqual((1, 2, 3), self.receivedRemoveEvents[0].values())
        
    def testClear_NoNotificationWhenNoItems(self):
        self.collection.clear()
        self.failIf(self.receivedRemoveEvents)
        
    def testModificationEventTypes(self):
        self.assertEqual([self.collection.addItemEventType(), 
                          self.collection.removeItemEventType()],
                         self.collection.modificationEventTypes())

        
class ObservableListTest(ObservableCollectionFixture, ObservableCollectionTestsMixin):
    def createObservableCollection(self):
        return patterns.ObservableList()

    def testAppendSameItemTwice(self):
        self.collection.append(1)
        self.collection.append(1)
        self.assertEqual(2, len(self.collection))
    

class ObservableSetTest(ObservableCollectionFixture, ObservableCollectionTestsMixin):
    def createObservableCollection(self):
        return patterns.ObservableSet()

    def testAppendSameItemTwice(self):
        self.collection.append(1)
        self.collection.append(1)
        self.assertEqual(1, len(self.collection))


class ListDecoratorTest_Constructor(test.TestCase):
    def testOriginalNotEmpty(self):
        observable = patterns.ObservableList([1, 2, 3])
        observer = patterns.ListDecorator(observable)
        self.assertEqual([1, 2, 3], observer)


class SetDecoratorTest_Constructor(test.TestCase):
    def testOriginalNotEmpty(self):
        observable = patterns.ObservableSet([1, 2, 3])
        observer = patterns.SetDecorator(observable)
        self.assertEqual([1, 2, 3], observer)


class ListDecoratorTest_AddItems(test.TestCase):
    def setUp(self):
        self.observable = patterns.ObservableList()
        self.observer = patterns.ListDecorator(self.observable)

    def testAppendToObservable(self):
        self.observable.append(1)
        self.assertEqual([1], self.observer)

    def testAppendToObserver(self):
        self.observer.append(1)
        self.assertEqual([1], self.observable)
        
    def testExtendObservable(self):
        self.observable.extend([1, 2, 3])
        self.assertEqual([1, 2, 3], self.observer)

    def testExtendObserver(self):
        self.observer.extend([1, 2, 3])
        self.assertEqual([1, 2, 3], self.observable)


class SetDecoratorTest_AddItems(test.TestCase):
    def setUp(self):
        self.observable = patterns.ObservableList()
        self.observer = patterns.SetDecorator(self.observable)

    def testAppendToObservable(self):
        self.observable.append(1)
        self.assertEqual([1], self.observer)

    def testAppendToObserver(self):
        self.observer.append(1)
        self.assertEqual([1], self.observable)
        
    def testExtendObservable(self):
        self.observable.extend([1, 2, 3])
        self.assertEqual([1, 2, 3], self.observer)

    def testExtendObserver(self):
        self.observer.extend([1, 2, 3])
        self.assertEqual([1, 2, 3], self.observable)


class ListDecoratorTest_RemoveItems(test.TestCase):
    def setUp(self):
        self.observable = patterns.ObservableList()
        self.observer = patterns.ListDecorator(self.observable)
        self.observable.extend([1, 2, 3])

    def testRemoveFromOriginal(self):
        self.observable.remove(1)
        self.assertEqual([2, 3], self.observer)

    def testRemoveFromObserver(self):
        self.observer.remove(1)
        self.assertEqual([2, 3], self.observable)

    def testRemoveItemsFromOriginal(self):
        self.observable.removeItems([1, 2])
        self.assertEqual([3], self.observer)

    def testRemoveItemsFromObserver(self):
        self.observer.removeItems([1, 2])
        self.assertEqual([3], self.observable)


class SetDecoratorTest_RemoveItems(test.TestCase):
    def setUp(self):
        self.observable = patterns.ObservableList()
        self.observer = patterns.SetDecorator(self.observable)
        self.observable.extend([1, 2, 3])

    def testRemoveFromOriginal(self):
        self.observable.remove(1)
        self.assertEqual([2, 3], self.observer)

    def testRemoveFromObserver(self):
        self.observer.remove(1)
        self.assertEqual([2, 3], self.observable)

    def testRemoveItemsFromOriginal(self):
        self.observable.removeItems([1, 2])
        self.assertEqual([3], self.observer)

    def testRemoveItemsFromObserver(self):
        self.observer.removeItems([1, 2])
        self.assertEqual([3], self.observable)
    

class ListDecoratorTest_ObserveTheObserver(test.TestCase):
    def setUp(self):
        self.list = patterns.ObservableList()
        self.observer = patterns.ListDecorator(self.list)
        patterns.Publisher().registerObserver(self.onAdd, 
            eventType=self.observer.addItemEventType(),
            eventSource=self.observer)
        patterns.Publisher().registerObserver(self.onRemove,
            eventType=self.observer.removeItemEventType(),
            eventSource=self.observer)
        self.receivedAddEvents = []
        self.receivedRemoveEvents = []

    def onAdd(self, event):
        self.receivedAddEvents.append(event)

    def onRemove(self, event):
        self.receivedRemoveEvents.append(event)

    def testExtendOriginal(self):
        self.list.extend([1, 2, 3])
        self.assertEqual((1, 2, 3), self.receivedAddEvents[0].values())

    def testExtendObserver(self):
        self.observer.extend([1, 2, 3])
        self.assertEqual((1, 2, 3), self.receivedAddEvents[0].values())

    def testRemoveItemsFromOriginal(self):
        self.list.extend([1, 2, 3])
        self.list.removeItems([1, 3])
        self.assertEqual((1, 3), self.receivedRemoveEvents[0].values())


class PublisherTest(test.TestCase):
    def setUp(self):
        self.publisher = patterns.Publisher()
        self.events = []
        self.events2 = []
        
    def onEvent(self, event):
        self.events.append(event)

    def onEvent2(self, event):
        self.events2.append(event)
                        
    def testPublisherIsSingleton(self):
        anotherPublisher = patterns.Publisher()
        self.failUnless(self.publisher is anotherPublisher)
        
    def testRegisterObserver(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType')
        self.assertEqual([self.onEvent], self.publisher.observers())
        
    def testRegisterObserver_Twice(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType')
        self.publisher.registerObserver(self.onEvent, eventType='eventType')
        self.assertEqual([self.onEvent], self.publisher.observers())
        
    def testRegisterObserver_ForTwoDifferentTypes(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType1')
        self.publisher.registerObserver(self.onEvent, eventType='eventType2')
        self.assertEqual([self.onEvent], self.publisher.observers())
        
    def testRegisterObserver_ListMethod(self):
        ''' A previous implementation of Publisher used sets. This caused a 
            "TypeError: list objects are unhashable" whenever one tried to use
            an instance method of a list (sub)class as callback. '''
        class List(list):
            def onEvent(self, *args):
                pass # pragma: no cover
        self.publisher.registerObserver(List().onEvent, eventType='eventType')   
         
    def testGetObservers_WithoutObservers(self):
        self.assertEqual([], self.publisher.observers())
        
    def testGetObserversForSpecificEventType_WithoutObservers(self):
        self.assertEqual([], self.publisher.observers(eventType='eventType'))

    def testGetObserversForSpecificEventType_WithObserver(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType')
        self.assertEqual([self.onEvent], 
            self.publisher.observers(eventType='eventType'))
            
    def testGetObserversForSpecificEventType_WhenDifferentTypesRegistered(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType1')
        self.publisher.registerObserver(self.onEvent, eventType='eventType2')
        self.assertEqual([self.onEvent], 
            self.publisher.observers(eventType='eventType1'))
            
    def testNotifyObservers_WithoutObservers(self):
        patterns.Event('eventType', self).send()
        self.failIf(self.events)

    def testNotifyObservers_WithObserverForDifferentEventType(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType1')
        patterns.Event('eventType2', self).send()
        self.failIf(self.events)
        
    def testNotifyObservers_WithObserverForRightEventType(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType')
        patterns.Event('eventType', self).send()
        self.assertEqual([patterns.Event('eventType', self)], self.events)
        
    def testNotifyObservers_WithObserversForSameAndDifferentEventTypes(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType1')
        self.publisher.registerObserver(self.onEvent, eventType='eventType2')
        patterns.Event('eventType1', self).send()
        self.assertEqual([patterns.Event('eventType1', self)], self.events)
        
    def testNotifyObservers_ForDifferentEventTypesWithOneEvent(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType1')
        self.publisher.registerObserver(self.onEvent2, eventType='eventType2')
        event = patterns.Event('eventType1', self)
        event.addSource(self, type='eventType2')
        event.send()
        self.assertEqual([patterns.Event('eventType1', self)], self.events)
        self.assertEqual([patterns.Event('eventType2', self)], self.events2)
        
    def testNotifyObserversWithEventWithoutTypes(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType')
        patterns.Event().send()
        self.failIf(self.events)

    def testNotifyObserversWithEventWithoutSources(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType')
        patterns.Event('eventType').send()
        self.failIf(self.events)
        
    def testRemoveObserverForAnyEventType_NotRegisteredBefore(self):
        self.publisher.removeObserver(self.onEvent)
        self.assertEqual([], self.publisher.observers())
        
    def testRemoveObserverForAnyEventType_RegisteredBefore(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType')
        self.publisher.removeObserver(self.onEvent)
        self.assertEqual([], self.publisher.observers())

    def testRemoveObserverForSpecificType_RegisteredForSameType(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType')
        self.publisher.removeObserver(self.onEvent, eventType='eventType')
        self.assertEqual([], self.publisher.observers())

    def testRemoveObserverForSpecificType_RegisteredForDifferentType(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType')
        self.publisher.removeObserver(self.onEvent, eventType='otherType')
        self.assertEqual([self.onEvent], self.publisher.observers())

    def testRemoveObserverForSpecificType_RegisteredForDifferentTypeThatHasObservers(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType')
        self.publisher.registerObserver(self.onEvent2, eventType='otherType')
        self.publisher.removeObserver(self.onEvent, eventType='otherType')
        self.assertEqual([self.onEvent], self.publisher.observers('eventType'))
        
    def testClear(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType')
        self.publisher.clear()
        self.assertEqual([], self.publisher.observers())        

    def testRegisterObserver_ForSpecificSource(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType', 
                                        eventSource='observable1')
        patterns.Event('eventType', 'observable2').send()
        self.failIf(self.events)
        
    def testNotifyObserver_ForSpecificSource(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType', 
                                        eventSource='observable1')
        event = patterns.Event('eventType', 'observable1')
        event.send()
        self.assertEqual([event], self.events)
        
    def testRemoveObserver_RegisteredForSpecificSource(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType', 
                                        eventSource='observable1')
        self.publisher.removeObserver(self.onEvent)
        event = patterns.Event('eventType', 'observable1')
        event.send()
        self.failIf(self.events)
        
    def testRemoveObserverForSpecificEventType_RegisteredForSpecificSource(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType', 
                                        eventSource='observable1')
        self.publisher.removeObserver(self.onEvent, eventType='eventType')
        patterns.Event('eventType', 'observable1').send() 
        self.failIf(self.events)

    def testRemoveObserverForSpecificEventSource(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType', 
                                        eventSource='observable1')
        self.publisher.registerObserver(self.onEvent, eventType='eventType',
                                        eventSource='observable2')
        self.publisher.removeObserver(self.onEvent, eventSource='observable1')
        patterns.Event('eventType', 'observable2').send()
        self.failUnless(self.events)
        
    def testRemoveObserverForSpecificEventTypeAndSource(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType1', 
                                        eventSource='observable1')
        self.publisher.registerObserver(self.onEvent, eventType='eventType1', 
                                        eventSource='observable2')
        self.publisher.registerObserver(self.onEvent, eventType='eventType2',
                                        eventSource='observable1')
        self.publisher.removeObserver(self.onEvent, eventType='eventType1',
                                      eventSource='observable1')
        patterns.Event('eventType1', 'observable1').send()
        self.failIf(self.events)
        patterns.Event('eventType2', 'observable1').send()
        self.failUnless(self.events)

    def testRemoveObserverForSpecificEventTypeAndSourceDoesNotRemoveOtherSources(self):
        self.publisher.registerObserver(self.onEvent, eventType='eventType1', 
                                        eventSource='observable1')
        self.publisher.registerObserver(self.onEvent, eventType='eventType1', 
                                        eventSource='observable2')
        self.publisher.registerObserver(self.onEvent, eventType='eventType2',
                                        eventSource='observable1')
        self.publisher.removeObserver(self.onEvent, eventType='eventType1',
                                      eventSource='observable1')
        patterns.Event('eventType1', 'observable2').send()
        self.failUnless(self.events)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import os
import urlparse
from taskcoachlib import patterns, mailer
from taskcoachlib.domain import base
from taskcoachlib.tools import openfile
from taskcoachlib.thirdparty.pubsub import pub
from taskcoachlib.domain.note.noteowner import NoteOwner


def getRelativePath(path, basePath=os.getcwd()):
    """Tries to guess the relative version of 'path' from 'basePath'. If
    not possible, return absolute 'path'. Both 'path' and 'basePath' must
    be absolute."""

    path = os.path.realpath(os.path.normpath(path))
    basePath = os.path.realpath(os.path.normpath(basePath))

    drive1, path1 = os.path.splitdrive(path)
    drive2, path2 = os.path.splitdrive(basePath)

    # No relative path is possible if the two are on different drives.
    if drive1 != drive2:
        return path

    if path1.startswith(path2):
        if path1 == path2:
            return ''

        if path2 == os.path.sep:
            return path1[1:].replace('\\', '/')

        return path1[len(path2) + 1:].replace('\\', '/')

    path1 = path1.split(os.path.sep)
    path2 = path2.split(os.path.sep)

    while path1 and path2 and path1[0] == path2[0]:
        path1.pop(0)
        path2.pop(0)

    while path2:
        path1.insert(0, '..')
        path2.pop(0)

    return os.path.join(*path1).replace('\\', '/')  # pylint: disable=W0142


class Attachment(base.Object, NoteOwner):
    ''' Abstract base class for attachments. '''

    type_ = 'unknown'

    def __init__(self, location, *args, **kwargs):
        if 'subject' not in kwargs:
            kwargs['subject'] = location
        super(Attachment, self).__init__(*args, **kwargs)
        self.__location = location

    def data(self):
        return None

    def setParent(self, parent):
        # FIXME: We shouldn't assume that pasted items are composite
        # in PasteCommand.
        pass

    def location(self):
        return self.__location

    def setLocation(self, location):
        if location != self.__location:
            self.__location = location
            self.markDirty()
            pub.sendMessage(self.locationChangedEventType(), newValue=location,
                            sender=self)

    @classmethod        
    def locationChangedEventType(class_):
        return 'pubsub.attachment.location'

    @classmethod
    def monitoredAttributes(class_):
        return base.Object.monitoredAttributes() + ['location']

    def open(self, workingDir=None):
        raise NotImplementedError

    def __cmp__(self, other):
        try:
            return cmp(self.location(), other.location())
        except AttributeError:
            return 1

    def __getstate__(self):
        try:
            state = super(Attachment, self).__getstate__()
        except AttributeError:
            state = dict()
        state.update(dict(location=self.location()))
        return state

    @patterns.eventSource
    def __setstate__(self, state, event=None):
        try:
            super(Attachment, self).__setstate__(state, event=event)
        except AttributeError:
            pass
        self.setLocation(state['location'])

    def __getcopystate__(self):
        return self.__getstate__()

    def __unicode__(self):
        return self.subject()
    
    @classmethod
    def modificationEventTypes(class_):
        eventTypes = super(Attachment, class_).modificationEventTypes()
        return eventTypes + [class_.locationChangedEventType()]


class FileAttachment(Attachment):
    type_ = 'file'

    def open(self, workingDir=None, openAttachment=openfile.openFile):  # pylint: disable=W0221
        return openAttachment(self.normalizedLocation(workingDir))

    def normalizedLocation(self, workingDir=None):
        location = self.location()
        if self.isLocalFile():
            if workingDir and not os.path.isabs(location):
                location = os.path.join(workingDir, location)
            location = os.path.normpath(location)
        return location

    def isLocalFile(self):
        return urlparse.urlparse(self.location())[0] == ''


class URIAttachment(Attachment):
    type_ = 'uri'

    def __init__(self, location, *args, **kwargs):
        if location.startswith('message:') and 'subject' not in kwargs:
            if self.settings.getboolean('os_darwin', 'getmailsubject'):
                subject = mailer.getSubjectOfMail(location[8:])
                if subject:
                    kwargs['subject'] = subject
            else:
                kwargs['subject'] = _('Mail.app message')
        super(URIAttachment, self).__init__(location, *args, **kwargs)

    def open(self, workingDir=None):
        return openfile.openFile(self.location())


class MailAttachment(Attachment):
    type_ = 'mail'

    def __init__(self, location, *args, **kwargs):
        self._readMail = kwargs.pop('readMail', mailer.readMail)
        subject, content = self._readMail(location)

        kwargs.setdefault('subject', subject)
        kwargs.setdefault('description', content)

        super(MailAttachment, self).__init__(location, *args, **kwargs)

    def open(self, workingDir=None):
        return mailer.openMail(self.location())

    def read(self):
        return self._readMail(self.location())

    def data(self):
        try:
            return file(self.location(), 'rb').read()
        except IOError:
            return None


def AttachmentFactory(location, type_=None, *args, **kwargs):
    if type_ is None:
        if location.startswith('URI:'):
            return URIAttachment(location[4:], subject=location[4:], description=location[4:])
        elif location.startswith('FILE:'):
            return FileAttachment(location[5:], subject=location[5:], description=location[5:])
        elif location.startswith('MAIL:'):
            return MailAttachment(location[5:], subject=location[5:], description=location[5:])

        return FileAttachment(location, subject=location, description=location)

    try:
        return {'file': FileAttachment,
                'uri': URIAttachment,
                'mail': MailAttachment}[type_](location, *args, **kwargs)
    except KeyError:
        raise TypeError('Unknown attachment type: %s' % type_)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import os, stat, atexit, tempfile
from taskcoachlib import patterns


class TempFiles(object):
    __metaclass__ = patterns.Singleton
    
    def __init__(self):
        self.__tempFiles = []
        atexit.register(self.cleanup)
        
    def register(self, filename):
        self.__tempFiles.append(filename)

    def cleanup(self):
        for name in self.__tempFiles:
            try:
                if os.name == 'nt':
                    os.chmod(name, stat.S_IREAD|stat.S_IWRITE)
                os.remove(name)
            except:
                pass # pylint: disable=W0702


def get_temp_file(**kwargs):
    ''' Return the name of a temporary file. This file will be registered
        for deletion at process termination. '''

    fd, filename = tempfile.mkstemp(**kwargs)
    os.close(fd)
    TempFiles().register(filename)
    return filename




"""
Test case to access the keyring from the command line
"""

import os.path

from keyring.tests.py30compat import unittest

import keyring.backend
from keyring import cli
from keyring import errors


class FakeKeyring(keyring.backend.KeyringBackend):
    PASSWORD = "GABUZOMEUH"

    def supported(self):
        return 1

    def set_password(self, service, username, password):
        pass

    def get_password(self, service, username):
        return self.PASSWORD

    def delete_password(self, service, username):
        pass

class SimpleKeyring(keyring.backend.KeyringBackend):
    """A very simple keyring"""

    def __init__(self):
        self.pwd = {}

    def supported(self):
        return 1

    def set_password(self, service, username, password):
        self.pwd[(service, username)] = password

    def get_password(self, service, username):
        try:
            return self.pwd[(service, username)]
        except KeyError:
            return None

    def delete_password(self, service, username):
        try:
            del self.pwd[(service, username)]
        except KeyError:
            raise errors.PasswordDeleteError("No key")

class CommandLineTestCase(unittest.TestCase):
    def setUp(self):
        self.old_keyring = keyring.get_keyring()

        self.cli = cli.CommandLineTool()
        self.cli.input_password = self.return_password
        self.cli.output_password = self.save_password
        self.cli.parser.error = self.mock_error
        self.cli.parser.print_help = lambda: None

        keyring.set_keyring(SimpleKeyring())

        self.password = ""
        self.password_returned = None
        self.last_error = None

    def tearDown(self):
        keyring.set_keyring(self.old_keyring)

    def return_password(self, *args, **kwargs):
        return self.password

    def save_password(self, password):
        self.password_returned = password

    def mock_error(self, error):
        self.last_error = error
        raise SystemExit()

    def test_wrong_arguments(self):
        self.assertEqual(1, self.cli.run([]))

        self.assertRaises(SystemExit, self.cli.run, ["get"])
        self.assertRaises(SystemExit, self.cli.run, ["get", "foo"])
        self.assertRaises(SystemExit, self.cli.run,
                          ["get", "foo", "bar", "baz"])

        self.assertRaises(SystemExit, self.cli.run, ["set"])
        self.assertRaises(SystemExit, self.cli.run, ["set", "foo"])
        self.assertRaises(SystemExit, self.cli.run,
                          ["set", "foo", "bar", "baz"])

        self.assertRaises(SystemExit, self.cli.run, ["foo", "bar", "baz"])

    def test_get_unexistent_password(self):
        self.assertEqual(1, self.cli.run(["get", "foo", "bar"]))
        self.assertEqual(None, self.password_returned)

    def test_set_and_get_password(self):
        self.password = "plop"
        self.assertEqual(0, self.cli.run(["set", "foo", "bar"]))
        self.assertEqual(0, self.cli.run(["get", "foo", "bar"]))
        self.assertEqual("plop", self.password_returned)

    def test_load_builtin_backend(self):
        self.assertEqual(1, self.cli.run([
            "get",
            "-b", "keyring.backends.file.PlaintextKeyring",
            "foo", "bar"]))
        backend = keyring.get_keyring()
        self.assertIsInstance(backend,
                              keyring.backends.file.PlaintextKeyring)

    def test_load_specific_backend_with_path(self):
        keyring_path = os.path.join(os.path.dirname(keyring.__file__), 'tests')
        self.assertEqual(0, self.cli.run(["get",
                                          "-b", "test_cli.FakeKeyring",
                                          "-p", keyring_path,
                                          "foo", "bar"]))

        # Somehow, this doesn't work, because the full dotted name of the class
        # is not the same as the one expected :(
        #self.assertIsInstance(keyring.get_keyring(), FakeKeyring)
        self.assertEqual(FakeKeyring.PASSWORD, self.password_returned)

    def test_load_wrong_keyrings(self):
        self.assertRaises(SystemExit, self.cli.run,
                          ["get", "foo", "bar",
                           "-b", "blablabla"  # ImportError
                          ])
        self.assertRaises(SystemExit, self.cli.run,
                          ["get", "foo", "bar",
                           "-b", "os.path.blabla"  # AttributeError
                          ])
        self.assertRaises(SystemExit, self.cli.run,
                          ["get", "foo", "bar",
                           "-b", "__builtin__.str"  # TypeError
                          ])


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(CommandLineTestCase))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest="test_suite")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from wxSchedule import *
from wxSchedulerConstants import *
from wxSchedulerPaint import *
import wx
import wxScheduleUtils as utils

if sys.version.startswith( "2.3" ):
	from sets import Set as set


class InvalidSchedule( Exception ):
	
	def __init__( self, value ):
		self.value = value
	
	def __str__( self ):
		return repr( self.value )
	

class wxSchedulerCore( wxSchedulerPaint ):
	
	def __init__( self, *args, **kwds ):
		self._currentDate = wx.DateTime.Now()
		self._weekstart	 = wxSCHEDULER_WEEKSTART_MONDAY
		
		super( wxSchedulerCore, self ).__init__( *args, **kwds )
		
		self._showOnlyWorkHour = True
		self._dc = None
		
		self._schedules = []
		self._schBind = []
		self._periodCount = 1
		
		#Internal (extenal?) init values
		self._startingHour = wx.DateTime().Now()
		self._endingHour = wx.DateTime().Now()
		self._startingPauseHour = wx.DateTime().Now()
		self._endingPauseHour = wx.DateTime().Now()
		
		self._startingHour.SetHour( 8 )
		self._startingHour.SetMinute( 00 )
		
		self._endingHour.SetHour( 18 )
		self._endingHour.SetMinute( 00 )
		
		self._startingPauseHour.SetHour( 13 )
		self._startingPauseHour.SetMinute( 00 )
		
		self._endingPauseHour.SetHour( 14 )
		self._endingPauseHour.SetMinute( 00 )
		
		self._calculateWorkHour()
		wxSchedulerCore.SetViewType( self )
			
	#-----------------------
	#   Internal methods
	#-----------------------
		
	def _calculateWorkHour( self ):
		""" 
		Do the calculation for work hour
		TO-DO: Make a better calculation!!
		"""
		
		# Update the current date according to 
		for i in ( self._startingHour, self._startingPauseHour,
							self._endingPauseHour, self._endingPauseHour ):
			i = utils.copyDate( self._currentDate )
		
		# Create the list
		self._lstDisplayedHours = []
		if self._showOnlyWorkHour:
			morningWorkTime = range( self._startingHour.GetHour(), self._startingPauseHour.GetHour() )
			afternoonWorkTime = range( self._endingPauseHour.GetHour(), self._endingHour.GetHour() )
			rangeWorkHour = morningWorkTime + afternoonWorkTime
		else:
			# Show all the hours
			rangeWorkHour = range( self._startingHour.GetHour(), self._endingHour.GetHour() )
		
		for H in rangeWorkHour:
			for M in ( 0, 30 ):
				hour = wx.DateTime().Now()
				hour = utils.copyDate( self._currentDate )
				hour.SetHour( H )
				hour.SetMinute( M )
				self._lstDisplayedHours.append( hour )
				
	def _calcRightDateOnMove( self, side ):
		""" 
		Calculate the right date when the user
		move the date on the next period
		"""
		
		if self._viewType == wxSCHEDULER_DAILY:
			offset = wx.DateSpan(days=1)
		elif self._viewType == wxSCHEDULER_WEEKLY:
			offset = wx.DateSpan(weeks=1)
		elif self._viewType == wxSCHEDULER_MONTHLY:
			daysAdd = self._currentDate.GetNumberOfDaysInMonth( self._currentDate.GetMonth() )
			offset = wx.DateSpan(months=1)

		if side == wxSCHEDULER_NEXT:
			self._currentDate.AddDS( offset )
		elif side == wxSCHEDULER_PREV:
			self._currentDate.SubtractDS( offset )
			
	#-----------------------
	#  External methods
	#-----------------------
	   
	def Add( self, schedules ):
		# Add schedules in list for visualization. Default is empty list
		# Call automatically Refresh() if at least one schedule is in range of 
		# current visualization
		if isinstance( schedules, wxSchedule ):
			self._schedules.append( schedules )
			
		elif isinstance( schedules, ( list, tuple ) ):
			#Control the schedule(s) passed
			for sc in schedules:
				if not isinstance( sc, wxSchedule ):
					raise InvalidSchedule, "Not a valid schedule"
				
				self._schedules.append( sc )
				
		else:
			raise ValueError( "Invalid value passed" )

		self.Refresh()
		
	def Delete( self, index ):
		""" Delete schedule in specified position or that specific wxSchedule
			Unbind also the event
		"""
		if isinstance( index, int ):
			schedule = self._schedules.pop( index )
		elif isinstance( index, wxSchedule ):
			#Remove the schedule from our list
			self._schedules.remove( index )
			schedule = index
		else:
			raise ValueError, "Passme only int or wxSchedule istances"
		
		#Remove from our bind list and unbind the event
		self._schBind.remove( schedule )
		schedule.Unbind( EVT_SCHEDULE_CHANGE )

		# Without that the object is never actually freed
		schedule.Destroy()

		self.Refresh()
		
	def DeleteAll( self ):
		"""
		Delete all schedules
		"""
		self.Freeze()
		try:
			while len( self._schedules ) > 0:
				self.Delete( 0 )
		finally:
			self.Thaw()

	def GetDate( self ):
		""" 
		Return the current day view
		If my view type is different than wxSCHEDULER_DAILY, I'll
		return the first day of the period
		"""
		return utils.copyDateTime( self._currentDate )
			
	def GetDc( self ):
		"""
		Return current DC
		"""
		return self._dc 
		
	def GetSchedules( self ):
		# Returns schedules in current days range. Useful for retrieve schedules 
		# in rendering mode
		return self._schedules

	def GetShowWorkHour( self ):
		#Return show work hour
		return self._showOnlyWorkHour
			
	def GetViewType( self ):
		# Return the current view type
		return self._viewType
	
	def GetWeekdayDate( self, weekday, day ):
		"""
		Return date of week day passed by
		Range values is 0 to 6
		"""
		if weekday == 6:
			weekday = 0
		else:
			weekday += 1
			
		date = day.GetWeekDayInSameWeek( weekday, self._weekstart )
		
		return utils.copyDateTime( date )
	
	def IsInRange( self, date=None, schedule=None ):
		"""
		Return True if the date or schedule is in the range of days displayed
		in current visualization type
		"""
		
		# Make the control
		if not ( date or schedule ) or not ( isinstance( date, ( wx.DateTime, wxSchedule ) ) or  
										  isinstance( schedule, wxSchedule ) ):
			raise ValueError, "Pass me at least one value"
		
		# Do a bad, but very useful hack that leave the developer pass an schedule at the first parameter
		if isinstance( date, wxSchedule ):
			schedule = date
			date = None
		
		# Create two new dates
		start = wx.DateTime.Now()
		end = wx.DateTime.Now()
		
		# Set the right, starting, parameters
		
		for DT, getPar in zip( ( start, end ), ( self._startingHour, self._endingHour ) ):
			DT.SetYear( self._currentDate.GetYear() )
			DT.SetMonth( self._currentDate.GetMonth() )
			DT.SetDay( self._currentDate.GetDay() )
			DT.SetHour( getPar.GetHour() )
			DT.SetMinute( getPar.GetMinute() )
		
		# Nothing to do
		if self._viewType == wxSCHEDULER_DAILY:
			pass
		# Set the right week
		elif self._viewType == wxSCHEDULER_WEEKLY:
			start = self.GetWeekdayDate( 0, start )
			end = self.GetWeekdayDate( 6, end )
		# End the end day
		elif self._viewType == wxSCHEDULER_MONTHLY:
			start.SetDay( 1 )
			end.SetDay( wx.DateTime.GetNumberOfDaysInMonth( end.GetMonth() ) )
		else:
			print "Why I'm here?"
			return
		
		# Make the control
		if date:
			return date.IsBetween( start, end )
		else:
			return start.IsEarlierThan( schedule.start ) and end.IsLaterThan( schedule.end )
		
	def Next( self, steps=1 ):
		self.Freeze()
		try:
			for step in xrange( steps ):
				self.SetViewType( wxSCHEDULER_NEXT )
		finally:
			self.Thaw()

	def Previous( self, steps=1 ):
		self.Freeze()
		try:
			for step in xrange( steps ):
				self.SetViewType( wxSCHEDULER_PREV )
		finally:
			self.Thaw()

	def SetDate( self, date=None ):
		# Go to the date. Default is today.
		if date == None:
			date = wx.DateTime.Now()
		self._currentDate = date
		self._calculateWorkHour()
		self.Refresh()
			
	def SetDc( self, dc=None ):
		"""
		Set alternative user defined DC.
		Call this method without arguments set automatically to wx.PaintDC on
		OnPaint()
		"""
		self._dc = dc

	def SetShowWorkHour( self, value ):
		#Set the show work hour value
		if not isinstance( value, ( bool, int ) ):
			raise ValueError, "Passme a bool at SetShowWorkHour"
		
		self._showOnlyWorkHour = value
		self._calculateWorkHour()
		self.InvalidateMinSize()
		self.Refresh()
	
	def SetViewType( self, view=None ):
		# Set the visualization type for the control. Default is daily
		if view == None:
			view = wxSCHEDULER_DAILY
			
		if not view in ( wxSCHEDULER_DAILY, wxSCHEDULER_WEEKLY,
							wxSCHEDULER_MONTHLY, wxSCHEDULER_TODAY,
							wxSCHEDULER_PREV, wxSCHEDULER_NEXT ):
				raise ValueError, "Pass me a valid view value"
		
		if view in ( wxSCHEDULER_TODAY, wxSCHEDULER_NEXT, wxSCHEDULER_PREV ):
			if view == wxSCHEDULER_TODAY:
				self._currentDate = wx.DateTime.Now()
			else: 
				self._calcRightDateOnMove( view )
			view = self._viewType
			
		self._viewType = view
		self._calculateWorkHour()
		
	def SetWeekStart( self, weekstart ):
		"""
		Set when the week start, on Monday or on Sunday
		Defaults it's start on Monday
		"""
		self._weekstart = weekstart
		self.Refresh()

	def GetWeekStart( self ):
		"""Returns the day the week starts."""
		return self._weekstart

	def SetWorkHours( self, start, stop ):
		"""
		Set start and end work hours
		"""
		self._startingHour.SetHour( max(0, start) )
		self._endingHour.SetHour( min(23, stop) )
		self._calculateWorkHour()
		self.InvalidateMinSize()
		self.Refresh()

	def SetPeriodCount( self, count ):
		"""
		Set the number of periods to display
		"""
		self._periodCount = count
		self.InvalidateMinSize()
		self.Refresh()

	def GetPeriodCount( self ):
		return self._periodCount



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import observer
import weakref


class Composite(object):
    def __init__(self, children=None, parent=None):
        super(Composite, self).__init__()
        self.__parent = parent if parent is None else weakref.ref(parent)
        self.__children = children or []
        for child in self.__children:
            child.setParent(self)
        
    def __getstate__(self):
        return dict(children=self.__children[:], 
                    parent=self.parent())
    
    def __setstate__(self, state):
        self.__parent = None if state['parent'] is None else weakref.ref(state['parent'])
        self.__children = state['children']

    def __getcopystate__(self):
        ''' Return the information needed to create a copy as a dict. '''
        try:
            state = super(Composite, self).__getcopystate__()
        except AttributeError:
            state = dict()
        state.update(dict(children=[child.copy() for child in self.__children], 
                          parent=self.parent()))
        return state
    
    def parent(self):
        return None if self.__parent is None else self.__parent()
    
    def ancestors(self):
        ''' Return the parent, and its parent, etc., as a list. '''
        parent = self.parent()
        return parent.ancestors() + [parent] if parent else []
    
    def family(self):
        ''' Return this object, its ancestors and all of its children 
            (recursively). '''
        return self.ancestors() + [self] + self.children(recursive=True)
    
    def setParent(self, parent):
        self.__parent = None if parent is None else weakref.ref(parent)

    def children(self, recursive=False):
        # Warning: this must satisfy the same condition as
        # allItemsSorted() below.

        if recursive:
            result = self.__children[:]
            for child in self.__children:
                result.extend(child.children(recursive=True))
            return result
        else:
            return self.__children
        
    def siblings(self, recursive=False):
        parent = self.parent()
        if parent:
            result = [child for child in parent.children() if child != self]
            if recursive:
                for child in result[:]:
                    result.extend(child.children(recursive=True))
            return result
        else:
            return []

    def copy(self, *args, **kwargs):
        kwargs['parent'] = self.parent()
        kwargs['children'] = [child.copy() for child in self.children()]
        return self.__class__(*args, **kwargs)
        
    def newChild(self, *args, **kwargs):
        kwargs['parent'] = self
        return self.__class__(*args, **kwargs)
    
    def addChild(self, child):
        self.__children.append(child)
        child.setParent(self)
        
    def removeChild(self, child):
        self.__children.remove(child)
        # We don't reset the parent of the child, because that makes restoring
        # the parent-child relationship easier.
        

class ObservableComposite(Composite):
    @observer.eventSource
    def __setstate__(self, state, event=None): # pylint: disable=W0221
        oldChildren = set(self.children())
        super(ObservableComposite, self).__setstate__(state)
        newChildren = set(self.children())
        childrenRemoved = oldChildren - newChildren
        # pylint: disable=W0142
        if childrenRemoved:
            self.removeChildEvent(event, *childrenRemoved)
        childrenAdded = newChildren - oldChildren
        if childrenAdded:
            self.addChildEvent(event, *childrenAdded)

    @observer.eventSource
    def addChild(self, child, event=None): # pylint: disable=W0221
        super(ObservableComposite, self).addChild(child)
        self.addChildEvent(event, child)

    def addChildEvent(self, event, *children):
        event.addSource(self, *children, **dict(type=self.addChildEventType()))

    @classmethod
    def addChildEventType(class_):
        return 'composite(%s).child.add'%class_

    @observer.eventSource
    def removeChild(self, child, event=None): # pylint: disable=W0221
        super(ObservableComposite, self).removeChild(child)
        self.removeChildEvent(event, child)

    def removeChildEvent(self, event, *children):    
        event.addSource(self, *children, **dict(type=self.removeChildEventType()))

    @classmethod
    def removeChildEventType(class_):
        return 'composite(%s).child.remove'%class_
    
    @classmethod
    def modificationEventTypes(class_):
        try:
            eventTypes = super(ObservableComposite, class_).modificationEventTypes()
        except AttributeError:
            eventTypes = []
        return eventTypes + [class_.addChildEventType(), 
                             class_.removeChildEventType()]
            

class CompositeCollection(object):
    def __init__(self, initList=None, *args, **kwargs):
        super(CompositeCollection, self).__init__(*args, **kwargs)
        self.extend(initList or [])
    
    def append(self, composite, event=None):
        return self.extend([composite], event=event)

    @observer.eventSource
    def extend(self, composites, event=None):
        if not composites:
            return
        compositesAndAllChildren = self._compositesAndAllChildren(composites) 
        super(CompositeCollection, self).extend(compositesAndAllChildren, 
                                                event=event)
        self._addCompositesToParent(composites, event)
            
    def _compositesAndAllChildren(self, composites):
        compositesAndAllChildren = set(composites) 
        for composite in composites:
            compositesAndAllChildren |= set(composite.children(recursive=True))
        return list(compositesAndAllChildren)

    def _addCompositesToParent(self, composites, event):
        for composite in composites:
            parent = composite.parent()
            if parent and parent in self and composite not in parent.children():
                parent.addChild(composite, event=event)
    
    def remove(self, composite, event=None):
        return self.removeItems([composite], event=event) if composite in self else event

    @observer.eventSource
    def removeItems(self, composites, event=None):
        if not composites:
            return
        compositesAndAllChildren = self._compositesAndAllChildren(composites)
        super(CompositeCollection, self).removeItems(compositesAndAllChildren, 
                                                     event=event)
        self._removeCompositesFromParent(composites, event)

    def _removeCompositesFromParent(self, composites, event):
        for composite in composites:
            parent = composite.parent()
            if parent:
                parent.removeChild(composite, event=event)
                            
    def rootItems(self):
        return [composite for composite in self if composite.parent() is None or \
                composite.parent() not in self]

    def allItemsSorted(self):
        """Returns a list of items and their children, so that if B is
        a child, direct or not, of A, then A will come first in the
        list."""

        result = []
        for item in self.rootItems():
            result.append(item)
            result.extend(item.children(recursive=True))
        return result


class CompositeSet(CompositeCollection, observer.ObservableSet):
    pass

    
class CompositeList(CompositeCollection, observer.ObservableList):
    pass




'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import operating_system
from wx.lib import masked
import wx
import locale


class FixOverwriteSelectionMixin(object):
    def _SetSelection(self, start, end):
        if operating_system.isGTK():  # pragma: no cover
            # By exchanging the start and end parameters we make sure that the 
            # cursor is at the start of the field so that typing overwrites the 
            # current field instead of moving to the next field:
            start, end = end, start
        super(FixOverwriteSelectionMixin, self)._SetSelection(start, end)

    def _OnKeyDown(self, event):
        # Allow keyboard navigation in notebook. Just skipping the event does not work;
        # propagate it all the way up...
        if event.GetKeyCode() == wx.WXK_TAB and event.GetModifiers() and hasattr(self.GetParent(), 'NavigateBook'):
            if self.GetParent().NavigateBook(event):
                return
        super(FixOverwriteSelectionMixin, self)._OnKeyDown(event)


class TextCtrl(FixOverwriteSelectionMixin, masked.TextCtrl):
    pass


class AmountCtrl(FixOverwriteSelectionMixin, masked.NumCtrl):
    def __init__(self, parent, value=0, locale_conventions=None):
        locale_conventions = locale_conventions or locale.localeconv()
        decimalChar = locale_conventions['decimal_point'] or '.'
        groupChar = locale_conventions['thousands_sep'] or ','
        groupDigits = len(locale_conventions['grouping']) > 1
        # The thousands separator may come up as ISO-8859-1 character
        # 0xa0, which looks like a space but isn't ASCII, which
        # confuses NumCtrl... Play it safe and avoid any non-ASCII
        # character here, or groupChars that consist of multiple characters.
        if len(groupChar) > 1 or ord(groupChar) >= 128:
            groupChar = ','
        # Prevent decimalChar and groupChar from being the same:
        if groupChar == decimalChar:
            groupChar = '.' if decimalChar == ',' else ','
        super(AmountCtrl, self).__init__(parent, value=value, 
            allowNegative=False, fractionWidth=2, selectOnEntry=True, 
            decimalChar=decimalChar, groupChar=groupChar, 
            groupDigits=groupDigits)
        

class TimeDeltaCtrl(TextCtrl):
    ''' Masked edit control for entering or displaying time deltas of the
        form <hour>:<minute>:<second>. Entering negative time deltas is not
        allowed, displaying negative time deltas is allowed if the control
        is read only. '''
    def __init__(self, parent, hours, minutes, seconds, readonly=False, 
                 negative_value=False, *args, **kwargs):
        # If the control is read only (meaning it could potentially have to 
        # show negative values) or if the value is actually negative, allow
        # the minus sign in the mask. Otherwise only allow for numbers.
        mask = 'X{9}:##:##' if negative_value or readonly else '#{9}:##:##'
        hours = self.__hour_string(hours, negative_value)
        super(TimeDeltaCtrl, self).__init__(parent, mask=mask, formatcodes='FS',
            fields=[masked.Field(formatcodes='Rr', defaultValue=hours),
                    masked.Field(defaultValue='%02d' % minutes),
                    masked.Field(defaultValue='%02d' % seconds)], 
            *args, **kwargs)
               
    def set_value(self, hours, minutes, seconds, negative_value=False):
        hours = self.__hour_string(hours, negative_value)
        self.SetCtrlParameters(formatcodes='FS',
            fields=[masked.Field(formatcodes='Rr', defaultValue=hours),
                    masked.Field(defaultValue='%02d' % minutes),
                    masked.Field(defaultValue='%02d' % seconds)])
        self.Refresh()
        
    @staticmethod
    def __hour_string(hours, negative_value):
        ''' If the value is negative (e.g. over budget), place a minus sign 
            before the hours number and make sure the field has the appropriate
            width. '''
        return '%9s' % ('-' + '%d' % hours) if negative_value else \
               '%9d' % hours



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import patterns


class Singleton(object):
    __metaclass__ = patterns.Singleton


class SingletonTest(test.TestCase):
    def tearDown(self):
        super(SingletonTest, self).tearDown()
        self.resetSingleton()
        
    def resetSingleton(self):
        Singleton.deleteInstance() # pylint: disable=E1101
        
    def testCreation(self):
        singleton = Singleton()
        self.failUnless(isinstance(singleton, Singleton))

    def testCreateTwice(self):
        single1 = Singleton()
        single2 = Singleton()
        self.failUnless(single1 is single2)

    def testSingletonsCanHaveInit(self):
        class SingletonWithInit:
            __metaclass__ = patterns.Singleton
            def __init__(self):
                self.a = 1
        single = SingletonWithInit()
        self.assertEqual(1, single.a)

    def testSingletonInitCanHaveArgs(self):
        class SingletonWithInit:
            __metaclass__ = patterns.Singleton
            def __init__(self, arg):
                self.a = arg
        single = SingletonWithInit('Yo')
        self.assertEqual('Yo', single.a)

    def testSingletonInitIsOnlyCalledOnce(self):
        class SingletonWithInit:
            _count = 0
            __metaclass__ = patterns.Singleton
            def __init__(self):
                SingletonWithInit._count += 1
        SingletonWithInit()
        SingletonWithInit()
        self.assertEqual(1, SingletonWithInit._count) # pylint: disable=W0212
        
    def testDeleteInstance(self):
        singleton1 = Singleton()
        self.resetSingleton()
        singleton2 = Singleton()
        self.failIf(singleton1 is singleton2)
        
    def testSingletonHasNoInstanceBeforeFirstCreation(self):
        self.failIf(Singleton.hasInstance()) # pylint: disable=E1101
        
    def testSingletonHasInstanceAfterFirstCreation(self):
        Singleton()
        self.failUnless(Singleton.hasInstance()) # pylint: disable=E1101
        
    def testSingletonHasInstanceAfterSecondCreation(self):
        Singleton()
        Singleton()
        self.failUnless(Singleton.hasInstance()) # pylint: disable=E1101
        
    def testSingletonHasNoInstanceAfterDeletion(self):
        Singleton()
        self.resetSingleton()
        self.failIf(Singleton.hasInstance()) # pylint: disable=E1101


class SingletonSubclassTest(test.TestCase):
    def testSubclassesAreSingletonsToo(self):
        class Sub(Singleton):
            pass
        sub1 = Sub()
        sub2 = Sub()
        self.failUnless(sub1 is sub2)

    def testDifferentSubclassesAreNotTheSameSingleton(self):
        class Sub1(Singleton):
            pass
        sub1 = Sub1()
        class Sub2(Singleton):
            pass
        sub2 = Sub2()
        self.failIf(sub1 is sub2)

    def testSubclassesCanHaveInit(self):
        class Sub(Singleton):
            def __init__(self):
                super(Sub, self).__init__()
                self.a = 1
        sub = Sub()
        self.assertEqual(1, sub.a)

    def testSubclassInitCanHaveArgs(self):
        class Sub(Singleton):
            def __init__(self, arg):
                super(Sub, self).__init__()
                self.arg = arg
        self.assertEqual('Yo', Sub('Yo').arg)

    def testSubclassInitIsOnlyCalledOnce(self):
        class Sub(Singleton):
            _count = 0
            def __init__(self):
                super(Sub, self).__init__()
                Sub._count += 1
        Sub()
        Sub()
        self.assertEqual(1, Sub._count) # pylint: disable=W0212




# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns
from taskcoachlib.domain.attribute import icon
from taskcoachlib.domain.date import DateTime, Now
from taskcoachlib.thirdparty.pubsub import pub
import attribute
import functools
import uuid
import re


class SynchronizedObject(object):
    STATUS_NONE    = 0
    STATUS_NEW     = 1
    STATUS_CHANGED = 2
    STATUS_DELETED = 3

    def __init__(self, *args, **kwargs):
        self.__status = kwargs.pop('status', self.STATUS_NEW)
        super(SynchronizedObject, self).__init__(*args, **kwargs)

    @classmethod
    def markDeletedEventType(class_):
        return 'object.markdeleted'

    @classmethod
    def markNotDeletedEventType(class_):
        return 'object.marknotdeleted'
        
    def __getstate__(self):
        try:
            state = super(SynchronizedObject, self).__getstate__()
        except AttributeError:
            state = dict()

        state['status'] = self.__status
        return state

    @patterns.eventSource
    def __setstate__(self, state, event=None):
        try:
            super(SynchronizedObject, self).__setstate__(state, event=event)
        except AttributeError:
            pass
        if state['status'] != self.__status:
            if state['status'] == self.STATUS_CHANGED:
                self.markDirty(event=event)
            elif state['status'] == self.STATUS_DELETED:
                self.markDeleted(event=event)
            elif state['status'] == self.STATUS_NEW:
                self.markNew(event=event)
            elif state['status'] == self.STATUS_NONE:
                self.cleanDirty(event=event)

    def getStatus(self):
        return self.__status
        
    @patterns.eventSource
    def markDirty(self, force=False, event=None):
        if not self.setStatusDirty(force):
            return
        event.addSource(self, self.__status, 
                        type=self.markNotDeletedEventType())

    def setStatusDirty(self, force=False):
        oldStatus = self.__status
        if self.__status == self.STATUS_NONE or force:
            self.__status = self.STATUS_CHANGED
            return oldStatus == self.STATUS_DELETED
        else:
            return False

    @patterns.eventSource
    def markNew(self, event=None):
        if not self.setStatusNew():
            return
        event.addSource(self, self.__status,
                        type=self.markNotDeletedEventType())
            
    def setStatusNew(self):
        oldStatus = self.__status
        self.__status = self.STATUS_NEW
        return oldStatus == self.STATUS_DELETED

    @patterns.eventSource
    def markDeleted(self, event=None):
        self.setStatusDeleted()
        event.addSource(self, self.__status, type=self.markDeletedEventType())

    def setStatusDeleted(self):
        self.__status = self.STATUS_DELETED

    @patterns.eventSource
    def cleanDirty(self, event=None):
        if not self.setStatusNone():
            return
        event.addSource(self, self.__status, 
                        type=self.markNotDeletedEventType())
            
    def setStatusNone(self):
        oldStatus = self.__status
        self.__status = self.STATUS_NONE
        return oldStatus == self.STATUS_DELETED

    def isNew(self):
        return self.__status == self.STATUS_NEW

    def isModified(self):
        return self.__status == self.STATUS_CHANGED

    def isDeleted(self):
        return self.__status == self.STATUS_DELETED

        
class Object(SynchronizedObject):
    rx_attributes = re.compile(r'\[(\w+):(.+)\]')

    def __init__(self, *args, **kwargs):
        Attribute = attribute.Attribute
        self.__creationDateTime = kwargs.pop('creationDateTime', None) or Now()
        self.__modificationDateTime = kwargs.pop('modificationDateTime', 
                                                 DateTime.min)
        self.__subject = Attribute(kwargs.pop('subject', ''), self, 
                                   self.subjectChangedEvent)
        self.__description = Attribute(kwargs.pop('description', ''), self,
                                       self.descriptionChangedEvent)
        self.__fgColor = Attribute(kwargs.pop('fgColor', None), self, 
                                   self.appearanceChangedEvent)
        self.__bgColor = Attribute(kwargs.pop('bgColor', None), self,
                                   self.appearanceChangedEvent)
        self.__font = Attribute(kwargs.pop('font', None), self,
                                self.appearanceChangedEvent)
        self.__icon = Attribute(kwargs.pop('icon', ''), self,
                                self.appearanceChangedEvent)
        self.__selectedIcon = Attribute(kwargs.pop('selectedIcon', ''), self,
                                        self.appearanceChangedEvent)
        self.__ordering = Attribute(kwargs.pop('ordering', 0L), self, self.orderingChangedEvent)
        self.__id = kwargs.pop('id', None) or str(uuid.uuid1())
        super(Object, self).__init__(*args, **kwargs)
    
    def __repr__(self):
        return self.subject()

    def __getstate__(self):
        try:
            state = super(Object, self).__getstate__()
        except AttributeError:
            state = dict()
        state.update(dict(id=self.__id, 
                          creationDateTime=self.__creationDateTime,
                          modificationDateTime=self.__modificationDateTime,
                          subject=self.__subject.get(), 
                          description=self.__description.get(),
                          fgColor=self.__fgColor.get(),
                          bgColor=self.__bgColor.get(),
                          font=self.__font.get(),
                          icon=self.__icon.get(),
                          ordering=self.__ordering.get(),
                          selectedIcon=self.__selectedIcon.get()))
        return state
    
    @patterns.eventSource
    def __setstate__(self, state, event=None):
        try:
            super(Object, self).__setstate__(state, event=event)
        except AttributeError:
            pass
        self.__id = state['id']
        self.setSubject(state['subject'], event=event)
        self.setDescription(state['description'], event=event)
        self.setForegroundColor(state['fgColor'], event=event)
        self.setBackgroundColor(state['bgColor'], event=event)
        self.setFont(state['font'], event=event)
        self.setIcon(state['icon'], event=event)
        self.setSelectedIcon(state['selectedIcon'], event=event)
        self.setOrdering(state['ordering'], event=event)
        self.__creationDateTime = state['creationDateTime']
        # Set modification date/time last to overwrite changes made by the 
        # setters above
        self.__modificationDateTime = state['modificationDateTime']

    def __getcopystate__(self):
        ''' Return a dictionary that can be passed to __init__ when creating
            a copy of the object. 
            
            E.g. copy = obj.__class__(**original.__getcopystate__()) '''
        try:
            state = super(Object, self).__getcopystate__()
        except AttributeError:
            state = dict()
        # Note that we don't put the id and the creation date/time in the state 
        # dict, because a copy should get a new id and a new creation date/time
        state.update(dict(\
            subject=self.__subject.get(), description=self.__description.get(),
            fgColor=self.__fgColor.get(), bgColor=self.__bgColor.get(),
            font=self.__font.get(), icon=self.__icon.get(),
            selectedIcon=self.__selectedIcon.get(),
            ordering=self.__ordering.get()))
        return state
    
    def copy(self):
        return self.__class__(**self.__getcopystate__())

    @classmethod
    def monitoredAttributes(class_):
        return ['ordering', 'subject', 'description', 'appearance']
 
    # Id:
       
    def id(self):
        return self.__id

    # Custom attributes
    def customAttributes(self, sectionName):
        attributes = set()
        for line in self.description().split('\n'):
            match = self.rx_attributes.match(line.strip())
            if match and match.group(1) == sectionName:
                attributes.add(match.group(2))
        return attributes

    # Editing date/time:
    
    def creationDateTime(self):
        return self.__creationDateTime
    
    def modificationDateTime(self):
        return self.__modificationDateTime
    
    def setModificationDateTime(self, dateTime):
        self.__modificationDateTime = dateTime

    @staticmethod
    def modificationDateTimeSortFunction(**kwargs):
        return lambda item: item.modificationDateTime()

    @staticmethod
    def creationDateTimeSortFunction(**kwargs):
        return lambda item: item.creationDateTime()

    # Subject:
    
    def subject(self):
        return self.__subject.get()
    
    def setSubject(self, subject, event=None):
        self.__subject.set(subject, event=event)
        
    def subjectChangedEvent(self, event):
        event.addSource(self, self.subject(), type=self.subjectChangedEventType())
    
    @classmethod    
    def subjectChangedEventType(class_):
        return '%s.subject' % class_
    
    @staticmethod
    def subjectSortFunction(**kwargs):
        ''' Function to pass to list.sort when sorting by subject. '''
        if kwargs.get('sortCaseSensitive', False):
            return lambda item: item.subject()
        else:
            return lambda item: item.subject().lower()
        
    @classmethod
    def subjectSortEventTypes(class_):
        ''' The event types that influence the subject sort order. '''
        return (class_.subjectChangedEventType(),)

    # Ordering:

    def ordering(self):
        return self.__ordering.get()

    def setOrdering(self, ordering, event=None):
        self.__ordering.set(ordering, event=event)

    def orderingChangedEvent(self, event):
        event.addSource(self, self.ordering(), type=self.orderingChangedEventType())

    @classmethod
    def orderingChangedEventType(class_):
        return '%s.ordering'%class_

    @staticmethod
    def orderingSortFunction(**kwargs):
        return lambda item: item.ordering()

    @classmethod
    def orderingSortEventTypes(class_):
        return (class_.orderingChangedEventType(),)

    # Description:
    
    def description(self):
        return self.__description.get()
    
    def setDescription(self, description, event=None):
        self.__description.set(description, event=event)
        
    def descriptionChangedEvent(self, event):
        event.addSource(self, self.description(), 
                        type=self.descriptionChangedEventType())
            
    @classmethod    
    def descriptionChangedEventType(class_):
        return '%s.description'%class_

    @staticmethod
    def descriptionSortFunction(**kwargs):
        ''' Function to pass to list.sort when sorting by description. '''
        if kwargs.get('sortCaseSensitive', False):
            return lambda item: item.description()
        else:
            return lambda item: item.description().lower()
    
    @classmethod
    def descriptionSortEventTypes(class_):
        ''' The event types that influence the description sort order. '''
        return (class_.descriptionChangedEventType(),)
    
    # Color:
    
    def setForegroundColor(self, color, event=None):
        self.__fgColor.set(color, event=event)
    
    def foregroundColor(self, recursive=False): # pylint: disable=W0613
        # The 'recursive' argument isn't actually used here, but some
        # code assumes composite objects where there aren't. This is
        # the simplest workaround.
        return self.__fgColor.get()

    def setBackgroundColor(self, color, event=None):
        self.__bgColor.set(color, event=event)
        
    def backgroundColor(self, recursive=False): # pylint: disable=W0613
        # The 'recursive' argument isn't actually used here, but some
        # code assumes composite objects where there aren't. This is
        # the simplest workaround.
        return self.__bgColor.get()
    
    # Font:
    
    def font(self, recursive=False): # pylint: disable=W0613
        # The 'recursive' argument isn't actually used here, but some
        # code assumes composite objects where there aren't. This is
        # the simplest workaround.
        return self.__font.get()
    
    def setFont(self, font, event=None):
        self.__font.set(font, event=event)

    # Icons:

    def icon(self):
        return self.__icon.get()

    def setIcon(self, icon, event=None):
        self.__icon.set(icon, event=event)

    def selectedIcon(self):
        return self.__selectedIcon.get()

    def setSelectedIcon(self, selectedIcon, event=None):
        self.__selectedIcon.set(selectedIcon, event=event)
    
    # Event types:
    
    @classmethod
    def appearanceChangedEventType(class_):
        return '%s.appearance'%class_
    
    def appearanceChangedEvent(self, event):
        event.addSource(self, type=self.appearanceChangedEventType())
    
    @classmethod
    def modificationEventTypes(class_):
        try:
            eventTypes = super(Object, class_).modificationEventTypes()
        except AttributeError:
            eventTypes = []
        return eventTypes + [class_.subjectChangedEventType(),
                             class_.descriptionChangedEventType(),
                             class_.appearanceChangedEventType(),
                             class_.orderingChangedEventType()]


class CompositeObject(Object, patterns.ObservableComposite):
    def __init__(self, *args, **kwargs):
        self.__expandedContexts = set(kwargs.pop('expandedContexts', []))
        super(CompositeObject, self).__init__(*args, **kwargs)

    def __getcopystate__(self):
        state = super(CompositeObject, self).__getcopystate__()
        state.update(dict(expandedContexts=self.expandedContexts()))
        return state

    @classmethod
    def monitoredAttributes(class_):
        return Object.monitoredAttributes() + ['expandedContexts']

    # Subject:

    def subject(self, recursive=False): # pylint: disable=W0221
        subject = super(CompositeObject, self).subject()
        if recursive and self.parent():
            subject = u'%s -> %s'%(self.parent().subject(recursive=True), subject)
        return subject

    def subjectChangedEvent(self, event):
        super(CompositeObject, self).subjectChangedEvent(event)
        for child in self.children():
            child.subjectChangedEvent(event)

    @staticmethod
    def subjectSortFunction(**kwargs):
        ''' Function to pass to list.sort when sorting by subject. '''
        recursive = kwargs.get('treeMode', False)
        if kwargs.get('sortCaseSensitive', False):
            return lambda item: item.subject(recursive=recursive)
        else:
            return lambda item: item.subject(recursive=recursive).lower()
        
    # Description:
        
    def description(self, recursive=False): # pylint: disable=W0221,W0613
        # Allow for the recursive flag, but ignore it
        return super(CompositeObject, self).description()
        
    # Expansion state:

    # Note: expansion state is stored by context. A context is a simple string
    # identifier (without comma's) to distinguish between different contexts,
    # i.e. viewers. A composite object may be expanded in one context and
    # collapsed in another.
    
    def isExpanded(self, context='None'):
        ''' Returns a boolean indicating whether the composite object is 
            expanded in the specified context. ''' 
        return context in self.__expandedContexts

    def expandedContexts(self):
        ''' Returns a list of contexts where this composite object is 
            expanded. ''' 
        return list(self.__expandedContexts)
    
    def expand(self, expand=True, context='None', notify=True):
        ''' Expands (or collapses) the composite object in the specified 
            context. ''' 
        if expand == self.isExpanded(context):
            return
        if expand:
            self.__expandedContexts.add(context)
        else:
            self.__expandedContexts.discard(context)
        if notify:
            pub.sendMessage(self.expansionChangedEventType(), newValue=expand,
                            sender=self)

    @classmethod
    def expansionChangedEventType(cls):
        ''' The event type used for notifying changes in the expansion state
            of a composite object. '''
        return 'pubsub.%s.expandedContexts' % cls.__name__.lower()

    def expansionChangedEvent(self, event):
        event.addSource(self, type=self.expansionChangedEventType())

    # The ChangeMonitor expects this...
    @classmethod
    def expandedContextsChangedEventType(class_):
        return class_.expansionChangedEventType()

    # Appearance:

    def appearanceChangedEvent(self, event):
        super(CompositeObject, self).appearanceChangedEvent(event)
        # Assume that most of the times our children change appearance too
        for child in self.children():
            child.appearanceChangedEvent(event)

    def foregroundColor(self, recursive=False):
        myFgColor = super(CompositeObject, self).foregroundColor()
        if not myFgColor and recursive and self.parent():
            return self.parent().foregroundColor(recursive=True)
        else:
            return myFgColor
                
    def backgroundColor(self, recursive=False):
        myBgColor = super(CompositeObject, self).backgroundColor()
        if not myBgColor and recursive and self.parent():
            return self.parent().backgroundColor(recursive=True)
        else:
            return myBgColor
    
    def font(self, recursive=False):
        myFont = super(CompositeObject, self).font()
        if not myFont and recursive and self.parent():
            return self.parent().font(recursive=True)
        else:
            return myFont

    def icon(self, recursive=False):
        myIcon = super(CompositeObject, self).icon()
        if not recursive:
            return myIcon
        if not myIcon and self.parent():
            myIcon = self.parent().icon(recursive=True)
        return self.pluralOrSingularIcon(myIcon, native=super(CompositeObject, self).icon() == '')

    def selectedIcon(self, recursive=False):
        myIcon = super(CompositeObject, self).selectedIcon()
        if not recursive:
            return myIcon
        if not myIcon and self.parent():
            myIcon = self.parent().selectedIcon(recursive=True)
        return self.pluralOrSingularIcon(myIcon, native=super(CompositeObject, self).selectedIcon() == '')

    def pluralOrSingularIcon(self, myIcon, native=True):
        hasChildren = any(child for child in self.children() if not child.isDeleted())
        mapping = icon.itemImagePlural if hasChildren else icon.itemImageSingular
        # If the icon comes from the user settings, only pluralize it; this is probably
        # the Way of the Least Astonishment
        if native or hasChildren:
            return mapping.get(myIcon, myIcon)
        return myIcon
    
    # Event types:

    @classmethod
    def modificationEventTypes(class_):
        return super(CompositeObject, class_).modificationEventTypes() + \
            [class_.expansionChangedEventType()]

    # Override SynchronizedObject methods to also mark child objects

    @patterns.eventSource
    def markDeleted(self, event=None):
        super(CompositeObject, self).markDeleted(event=event)
        for child in self.children():
            child.markDeleted(event=event)

    @patterns.eventSource
    def markNew(self, event=None):
        super(CompositeObject, self).markNew(event=event)
        for child in self.children():
            child.markNew(event=event)

    @patterns.eventSource
    def markDirty(self, force=False, event=None):
        super(CompositeObject, self).markDirty(force, event=event)
        for child in self.children():
            child.markDirty(force, event=event)

    @patterns.eventSource            
    def cleanDirty(self, event=None):
        super(CompositeObject, self).cleanDirty(event=event)
        for child in self.children():
            child.cleanDirty(event=event)



'''
Task Coach - Your friendly task manager
Copyright (C) 2014 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
import datetime
import math


wxEVT_EVENT_SELECTION_CHANGED = wx.NewEventType()
EVT_EVENT_SELECTION_CHANGED = wx.PyEventBinder(wxEVT_EVENT_SELECTION_CHANGED)

wxEVT_EVENT_DATES_CHANGED = wx.NewEventType()
EVT_EVENT_DATES_CHANGED = wx.PyEventBinder(wxEVT_EVENT_DATES_CHANGED)


class _HitResult(object):
    HIT_START        = 0
    HIT_IN           = 1
    HIT_END          = 2

    def __init__(self, x, y, event, dateTime):
        self.x, self.y = x, y
        self.event = event
        self.dateTime = dateTime
        self.position = self.HIT_IN


class _Watermark(object):
    def __init__(self):
        self.__values = []

    def height(self, start, end):
        r = 0
        for (ints, inte, h) in self.__values:
            if not (end < ints or start >= inte):
                r = max(r, h)
        return r

    def totalHeight(self):
        return max([h for ints, inte, h in self.__values]) if self.__values else 0

    def add(self, start, end, h):
        self.__values.append((start, end, h))


def total_seconds(td): # Method new in 2.7
    return (1.0 * td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6


def shortenText(gc, text, maxW):
    shortText = text
    idx = len(text) // 2
    while True:
        w, h = gc.GetTextExtent(shortText)
        if w <= maxW:
            return shortText
        idx -= 1
        if idx == 0:
            return u'\u2026'
        shortText = text[:idx] + u'\u2026' + text[-idx:]


class CalendarCanvas(wx.Panel):
    _gradVal = 0.2

    MS_IDLE          = 0
    MS_HOVER_LEFT    = 1
    MS_HOVER_RIGHT   = 2
    MS_DRAG_LEFT     = 3
    MS_DRAG_RIGHT    = 4
    MS_DRAG_START    = 5
    MS_DRAGGING      = 6

    def __init__(self, parent, start=None, end=None):
        self._start = start or datetime.datetime.combine(datetime.datetime.now().date(), datetime.time(0, 0, 0))
        self._end = end or self._start + datetime.timedelta(days=7)
        super(CalendarCanvas, self).__init__(parent, wx.ID_ANY, style=wx.FULL_REPAINT_ON_RESIZE)

        self._coords = dict() # Event => (startIdx, endIdx, startIdxRecursive, endIdxRecursive, yMin, yMax)
        self._maxIndex = 0
        self._minSize = (0, 0)

        # Drawing attributes
        self._precision = 1 # Minutes
        self._gridSize = 15 # Minutes
        self._eventHeight = 32.0
        self._eventWidthMin = 0.1
        self._eventWidth = 0.1
        self._margin = 5.0
        self._marginTop = 22.0
        self._outlineColorDark = wx.Colour(180, 180, 180)
        self._outlineColorLight = wx.Colour(210, 210, 210)
        self._headerSpans = []
        self._daySpans = []
        self._selection = set()
        self._mouseState = self.MS_IDLE
        self._mouseOrigin = None
        self._mouseDragPos = None
        self._todayColor = wx.Colour(0, 0, 128)

        self._hScroll = wx.ScrollBar(self, wx.ID_ANY, style=wx.SB_HORIZONTAL)
        self._vScroll = wx.ScrollBar(self, wx.ID_ANY, style=wx.SB_VERTICAL)

        self._hScroll.Hide()
        self._vScroll.Hide()

        wx.EVT_SCROLL(self._hScroll, self._OnScroll)
        wx.EVT_SCROLL(self._vScroll, self._OnScroll)

        wx.EVT_PAINT(self, self._OnPaint)
        wx.EVT_SIZE(self, self._OnResize)
        wx.EVT_LEFT_DOWN(self, self._OnLeftDown)
        wx.EVT_LEFT_UP(self, self._OnLeftUp)
        wx.EVT_RIGHT_DOWN(self, self._OnRightDown)
        wx.EVT_MOTION(self, self._OnMotion)
        self._Invalidate()

    # Methods to override

    def IsWorked(self, date):
        return not date.isoweekday() in [6, 7]

    def FormatDateTime(self, dateTime):
        return dateTime.strftime('%A')

    def GetRootEvents(self):
        return list()

    def GetStart(self, event):
        raise NotImplementedError

    def GetEnd(self, event):
        raise NotImplementedError

    def GetText(self, event):
        raise NotImplementedError

    def GetChildren(self, event):
        raise NotImplementedError

    def GetBackgroundColor(self, event):
        raise NotImplementedError

    def GetForegroundColor(self, event):
        raise NotImplementedError

    def GetProgress(self, event):
        raise NotImplementedError

    def GetIcons(self, event):
        raise NotImplementedError

    def GetFont(self, event):
        raise NotImplementedError

    # Get/Set

    def GetPrecision(self):
        return self._precision

    def SetPrecision(self, precision):
        self._precision = precision
        self._Invalidate()
        self.Refresh()

    def GetEventHeight(self):
        return self._eventHeight

    def SetEventHeight(self, height):
        self._eventHeight = 1.0 * height
        self._Invalidate()
        self.Refresh()

    def GetEventWidth(self):
        return self._eventWidthMin

    def SetEventWidth(self, width):
        self._eventWidthMin = width
        self._Invalidate()
        self.Refresh()

    def GetMargin(self):
        return self._margin

    def SetMargin(self, margin):
        self._margin = 1.0 * margin
        self._Invalidate()
        self.Refresh()

    def OutlineColorDark(self):
        return self._outlineColorDark

    def SetOutlineColorDark(self, color):
        self._outlineColorDark = color
        self.Refresh()

    def OutlineColorLight(self):
        return self._outlineColorLight

    def SetOutlineColorLight(self, color):
        self._outlineColorLight = color
        self.Refresh()

    def TodayColor(self):
        return self._todayColor

    def SetTodayColor(self, color):
        self._todayColor = color
        self.Refresh()

    def ViewSpan(self):
        return (self._start, self._end)

    def SetViewSpan(self, start, end):
        self._start = start
        self._end = end
        self._Invalidate()
        self.Refresh()

    def Selection(self):
        return self._selection

    def Select(self, events):
        self._selection = set(events) & set(self._coords.keys())
        e = wx.PyCommandEvent(wxEVT_EVENT_SELECTION_CHANGED)
        e.selection = set(self._selection)
        e.SetEventObject(self)
        self.ProcessEvent(e)
        self.Refresh()

    def HitTest(self, x, y):
        w, h = self.GetClientSizeTuple()

        if y <= self._marginTop:
            return None
        if self._hScroll.IsShown():
            h -= self._hScroll.GetClientSizeTuple()[1]
            if y >= h:
                return None
        if self._vScroll.IsShown():
            w -= self._vScroll.GetClientSizeTuple()[0]
            if x >= w:
                return None

        if self._hScroll.IsShown():
            x += self._hScroll.GetThumbPosition()
        if self._vScroll.IsShown():
            y += self._vScroll.GetThumbPosition()

        xIndex = int(x / self._eventWidth)
        yIndex = int((y - self._marginTop) / (self._eventHeight + self._margin))
        dateTime = self._start + datetime.timedelta(minutes=self._precision * xIndex)

        for event, (startIndex, endIndex, startIndexRecursive, endIndexRecursive, yMin, yMax) in self._coords.items():
            if xIndex >= startIndexRecursive and xIndex < endIndexRecursive and yIndex >= yMin and yIndex < yMax:
                # May be a child
                children = []
                self._Flatten(event, children)
                for candidate in reversed(children):
                    if candidate in self._coords:
                        si, ei, sir, eir, ymin, ymax = self._coords[candidate]
                        if si is not None and abs(x - si * self._eventWidth) <= self._margin and yIndex >= ymin and yIndex < ymax:
                            result = _HitResult(x, y, candidate, dateTime)
                            result.position = result.HIT_START
                            return result
                        if ei is not None and abs(x - ei * self._eventWidth) <= self._margin and yIndex >= ymin and yIndex < ymax:
                            result = _HitResult(x, y, candidate, dateTime)
                            result.position = result.HIT_END
                            return result
                        if xIndex >= sir and xIndex < eir and yIndex >= ymin and yIndex < ymax:
                            result = _HitResult(x, y, candidate, dateTime)
                            result.position = result.HIT_IN
                            return result
                # Since the list contains at least 'event'...
                assert 0

        # We didn't hit any event.
        result = _HitResult(x, y, None, dateTime)
        return result

    def _Flatten(self, event, result):
        result.append(event)
        for child in self.GetChildren(event):
            self._Flatten(child, result)

    def _DrawEvent(self, gc, event):
        if event in self._coords:
            startIndex, endIndex, startIndexRecursive, endIndexRecursive, yMin, yMax = self._coords[event]
            if self.GetChildren(event):
                self._DrawParent(gc, startIndex, endIndex, startIndexRecursive, endIndexRecursive, yMin, yMax, event, self._eventWidth)
            else:
                self._DrawLeaf(gc, startIndex, endIndex, yMin, yMax, event, self._eventWidth)
        for child in self.GetChildren(event):
            self._DrawEvent(gc, child)

    def _OnPaint(self, event):
        w, h = self.GetClientSizeTuple()
        vw = max(w, self._minSize[0])
        vh = max(h, self._minSize[1])
        dx = dy = 0
        if self._hScroll.IsShown():
            vh -= self._hScroll.GetClientSizeTuple()[1]
            dx = self._hScroll.GetThumbPosition()
        if self._vScroll.IsShown():
            vw -= self._vScroll.GetClientSizeTuple()[0]
            dy = self._vScroll.GetThumbPosition()

        bmp = wx.EmptyBitmap(vw, vh)
        memDC = wx.MemoryDC()
        memDC.SelectObject(bmp)
        try:
            memDC.SetBackground(wx.WHITE_BRUSH)
            memDC.Clear()
            gc = wx.GraphicsContext.Create(memDC)
            self._Draw(gc, vw, vh, dx, dy)
            dc = wx.PaintDC(self)
            dc.Blit(0, 0, vw, vh, memDC, 0, 0)
        finally:
            memDC.SelectObject(wx.NullBitmap)

    def _Draw(self, gc, vw, vh, dx, dy):
        gc.PushState()
        try:
            gc.Translate(-dx, 0.0)
            self._DrawHeader(gc, vw, vh)
        finally:
            gc.PopState()

        gc.PushState()
        try:
            gc.Translate(-dx, -dy)
            gc.Clip(0, self._marginTop + dy, vw, vh)
            for event in self.GetRootEvents():
                self._DrawEvent(gc, event)
            self._DrawNow(gc, vh + dy)
            self._DrawDragImage(gc)
        finally:
            gc.PopState()

    def _DrawHeader(self, gc, w, h):
        gc.SetPen(wx.Pen(self._outlineColorDark))
        for startIndex, endIndex in self._daySpans:
            date = (self._start + datetime.timedelta(minutes=self._precision * startIndex)).date()
            x0 = startIndex * self._eventWidth
            x1 = endIndex * self._eventWidth
            if date == datetime.datetime.now().date():
                gc.SetBrush(self._Gradient(gc, self._todayColor, x0, self._marginTop, x1 - x0, h))
            elif self.IsWorked(date):
                gc.SetBrush(wx.WHITE_BRUSH)
            else:
                gc.SetBrush(self._Gradient(gc, self._outlineColorDark, x0, self._marginTop, x1 - x0, h))
            gc.DrawRectangle(x0, self._marginTop, x1 - x0, h)

        gc.SetFont(wx.NORMAL_FONT, wx.BLACK)
        gc.SetPen(wx.Pen(self._outlineColorDark))
        for startIndex, endIndex in self._headerSpans:
            x0 = startIndex * self._eventWidth
            x1 = endIndex * self._eventWidth
            gc.SetBrush(self._Gradient(gc, self._outlineColorLight, x0, 0, x1 - x0, self._marginTop - 2.0))
            gc.DrawRectangle(x0, 0, x1 - x0, self._marginTop - 2.0)
            text = shortenText(gc, self.FormatDateTime(self._start + datetime.timedelta(minutes=self._precision * startIndex)), x1 - x0)
            tw, th = gc.GetTextExtent(text)
            gc.DrawText(text, x0 + (x1 - x0 - tw) / 2, (self._marginTop - 2.0 - th) / 2)

    def _DrawNow(self, gc, h):
        now = datetime.datetime.now()
        x = int((now - self._start).total_seconds() / 60.0 / self._precision * self._eventWidth) - 0.5
        gc.SetPen(wx.Pen(wx.Colour(0, 128, 0)))
        gc.SetBrush(wx.Brush(wx.Colour(0, 128, 0)))

        path = gc.CreatePath()
        path.MoveToPoint(x - 4.0, self._marginTop)
        path.AddLineToPoint(x + 4.0, self._marginTop)
        path.AddLineToPoint(x, self._marginTop + 4.0)
        path.AddLineToPoint(x, h + self._marginTop)
        path.AddLineToPoint(x, self._marginTop + 4.0)
        path.CloseSubpath()
        gc.DrawPath(path)

    def _DrawDragImage(self, gc):
        if self._mouseDragPos is not None:
            if self._mouseState in [self.MS_DRAG_LEFT, self.MS_DRAG_RIGHT]:
                d1 = self._mouseDragPos
                d2 = self.GetEnd(self._mouseOrigin.event) if self._mouseState == self.MS_DRAG_LEFT else self.GetStart(self._mouseOrigin.event)
                d1, d2 = min(d1, d2), max(d1, d2)

                x0 = int(total_seconds(d1 - self._start) / 60 / self._precision) * self._eventWidth
                x1 = int(total_seconds(d2 - self._start) / 60 / self._precision) * self._eventWidth
                y0 = self._coords[self._mouseOrigin.event][4] * (self._eventHeight + self._margin) + self._marginTop
                y1 = self._coords[self._mouseOrigin.event][5] * (self._eventHeight + self._margin) + self._marginTop - self._margin

                gc.SetBrush(wx.Brush(wx.Colour(0, 0, 128, 128)))
                gc.DrawRoundedRectangle(x0, y0, x1 - x0, y1 - y0, 5.0)

                gc.SetFont(wx.NORMAL_FONT, wx.RED)
                text = self._mouseDragPos.strftime('%c')
                tw, th = gc.GetTextExtent(text)
                if self._mouseState == self.MS_DRAG_LEFT:
                    tx = x0 + self._margin
                elif self._mouseState == self.MS_DRAG_RIGHT:
                    tx = x1 - self._margin - tw
                ty = y0 + (y1 - y0 - th) / 2
                gc.DrawText(text, tx, ty)
            elif self._mouseState == self.MS_DRAGGING:
                x0 = int(total_seconds(self._mouseDragPos - self._start) / 60 / self._precision) * self._eventWidth
                x1 = int(total_seconds(self._mouseDragPos + (self.GetEnd(self._mouseOrigin.event) - self.GetStart(self._mouseOrigin.event)) - self._start) / 60 / self._precision) * self._eventWidth
                y0 = self._coords[self._mouseOrigin.event][4] * (self._eventHeight + self._margin) + self._marginTop
                y1 = self._coords[self._mouseOrigin.event][5] * (self._eventHeight + self._margin) + self._marginTop - self._margin

                gc.SetBrush(wx.Brush(wx.Colour(0, 0, 128, 128)))
                gc.DrawRoundedRectangle(x0, y0, x1 - x0, y1 - y0, 5.0)

                gc.SetFont(wx.NORMAL_FONT, wx.RED)
                text = u'%s -> %s' % (self._mouseDragPos.strftime('%c'), (self._mouseDragPos + (self.GetEnd(self._mouseOrigin.event) - self.GetStart(self._mouseOrigin.event))).strftime('%c'))
                tw, th = gc.GetTextExtent(text)
                gc.DrawText(text, x0 + (x1 - x0 - tw) / 2, y0 + (y1 - y0 - th) / 2)


    def _GetCursorDate(self):
        x, y = self.ScreenToClientXY(*wx.GetMousePosition())
        if self._hScroll.IsShown():
            x += self._hScroll.GetThumbPosition()
        return self._start + datetime.timedelta(minutes=int(self._precision * x / self._eventWidth))

    def _OnResize(self, event=None):
        if event is None:
            w, h = self.GetClientSizeTuple()
        else:
            w, h = event.GetSize()

        _, hh = self._hScroll.GetClientSizeTuple()
        vw, _ = self._vScroll.GetClientSizeTuple()

        self._hScroll.SetDimensions(0, h - hh, w - vw, hh)
        self._vScroll.SetDimensions(w - vw, self._marginTop, vw, h - hh - self._marginTop)

        minW, minH = self._minSize

        # Not perfect, but it will do.
        if w - vw < minW:
            self._hScroll.SetScrollbar(self._hScroll.GetThumbPosition(), w - vw, minW, w - vw, True)
            self._hScroll.Show()
            h -= hh
        else:
            self._hScroll.Hide()

        if h - hh - self._marginTop < minH:
            self._vScroll.SetScrollbar(self._vScroll.GetThumbPosition(), h - hh - self._marginTop, minH, h - hh - self._marginTop, True)
            self._vScroll.Show()
            w -= vw
        else:
            self._vScroll.Hide()

        self._eventWidth = max(self._eventWidthMin, 1.0 * max(w, minW) / self._maxIndex)

        if event is not None:
            event.Skip()

    def _OnLeftDown(self, event):
        result = self.HitTest(event.GetX(), event.GetY())
        if result is None:
            return

        if self._mouseState == self.MS_IDLE:
            changed = False
            if result.event is None:
                if self._selection:
                    changed = True
                    self._selection = set()
                    self.Refresh()
            else:
                if event.ShiftDown():
                    events = []
                    self._Flatten(result.event, events)
                else:
                    events = [result.event]
                events = set(events) & set(self._coords.keys())

                if event.CmdDown():
                    for e in events:
                        if e in self._selection:
                            self._selection.remove(e)
                            changed = True
                        else:
                            self._selection.add(e)
                            changed = True
                else:
                    if self._selection != events:
                        changed = True
                        self._selection = events

                if result.position == result.HIT_IN:
                    self._mouseOrigin = result
                    self._mouseState = self.MS_DRAG_START
                self.Refresh()

            if changed:
                e = wx.PyCommandEvent(wxEVT_EVENT_SELECTION_CHANGED)
                e.selection = set(self._selection)
                e.SetEventObject(self)
                self.ProcessEvent(e)
        elif self._mouseState in [self.MS_HOVER_LEFT, self.MS_HOVER_RIGHT]:
            self.CaptureMouse()
            self._mouseState += self.MS_DRAG_LEFT - self.MS_HOVER_LEFT

    def _OnLeftUp(self, event):
        if self._mouseState in [self.MS_DRAG_LEFT, self.MS_DRAG_RIGHT]:
            self.ReleaseMouse()
            wx.SetCursor(wx.NullCursor)

            e = wx.PyCommandEvent(wxEVT_EVENT_DATES_CHANGED)
            e.event = self._mouseOrigin.event
            e.start = self._mouseDragPos if self._mouseState == self.MS_DRAG_LEFT else self.GetStart(self._mouseOrigin.event)
            e.end = self._mouseDragPos if self._mouseState == self.MS_DRAG_RIGHT else self.GetEnd(self._mouseOrigin.event)
            e.SetEventObject(self)
            self.ProcessEvent(e)
        elif self._mouseState == self.MS_DRAGGING:
            self.ReleaseMouse()
            wx.SetCursor(wx.NullCursor)

            e = wx.PyCommandEvent(wxEVT_EVENT_DATES_CHANGED)
            e.event = self._mouseOrigin.event
            e.start = self._mouseDragPos
            e.end = e.start + (self.GetEnd(self._mouseOrigin.event) - self.GetStart(self._mouseOrigin.event))
            e.SetEventObject(self)
            self.ProcessEvent(e)

        self._mouseState = self.MS_IDLE
        self._mouseOrigin = None
        self._mouseDragPos = None
        self.Refresh()

    def _OnRightDown(self, event):
        result = self.HitTest(event.GetX(), event.GetY())
        if result is None:
            return

        changed = False
        if result.event is None:
            if self._selection:
                self._selection = set()
                changed = True
                self.Refresh()
        else:
            if result.event not in self._selection:
                self._selection = set([result.event])
                changed = True
                self.Refresh()

        if changed:
            e = wx.PyCommandEvent(wxEVT_EVENT_SELECTION_CHANGED)
            e.selection = set(self._selection)
            e.SetEventObject(self)
            self.ProcessEvent(e)

    def _OnMotion(self, event):
        result = self.HitTest(event.GetX(), event.GetY())

        if result is not None:
            if self._mouseState == self.MS_IDLE:
                if result.event is not None and result.position in [result.HIT_START, result.HIT_END]:
                    self._mouseOrigin = result
                    self._mouseState = self.MS_HOVER_LEFT if result.position == result.HIT_START else self.MS_HOVER_RIGHT
                    wx.SetCursor(wx.StockCursor(wx.CURSOR_SIZEWE))
            elif self._mouseState in [self.MS_HOVER_LEFT, self.MS_HOVER_RIGHT]:
                if result.event is None or result.position not in [result.HIT_START, result.HIT_END]:
                    self._mouseOrigin = None
                    self._mouseDragPos = None
                    self._mouseState = self.MS_IDLE
                    wx.SetCursor(wx.NullCursor)

        if self._mouseState in [self.MS_DRAG_LEFT, self.MS_DRAG_RIGHT]:
            dateTime = self._GetCursorDate()
            precision = self._gridSize if event.ShiftDown() else self._precision
            if self._mouseState == self.MS_DRAG_LEFT:
                dateTime = self._start + datetime.timedelta(seconds=math.floor(total_seconds(dateTime - self._start) / 60 / precision) * precision * 60)
                dateTime = min(self.GetEnd(self._mouseOrigin.event) - datetime.timedelta(minutes=precision), dateTime)
            if self._mouseState == self.MS_DRAG_RIGHT:
                dateTime = self._start + datetime.timedelta(seconds=math.ceil(total_seconds(dateTime - self._start) / 60 / precision) * precision * 60)
                dateTime = max(self.GetStart(self._mouseOrigin.event) + datetime.timedelta(minutes=precision), dateTime)
            self._mouseDragPos = dateTime

            self.Refresh()
        elif self._mouseState == self.MS_DRAG_START:
            if self.GetStart(self._mouseOrigin.event) is not None and self.GetEnd(self._mouseOrigin.event) is not None:
                dx = abs(event.GetX() - self._mouseOrigin.x)
                dy = abs(event.GetY() - self._mouseOrigin.y)
                if dx > wx.SystemSettings.GetMetric(wx.SYS_DRAG_X) / 2 or dy > wx.SystemSettings.GetMetric(wx.SYS_DRAG_Y) / 2:
                    self.CaptureMouse()
                    wx.SetCursor(wx.StockCursor(wx.CURSOR_HAND))
                    self._mouseState = self.MS_DRAGGING
                    self.Refresh()
        elif self._mouseState == self.MS_DRAGGING:
            dx = event.GetX() - self._mouseOrigin.x
            precision = self._gridSize if event.ShiftDown() else self._precision
            delta = datetime.timedelta(minutes=math.floor(dx / self._eventWidth * self._precision / precision) * precision)
            self._mouseDragPos = self.GetStart(self._mouseOrigin.event) + delta
            self.Refresh()

    def _OnScroll(self, event):
        self.Refresh()
        event.Skip()

    def _Gradient(self, gc, color, x, y, w, h):
        r = color.Red()
        g = color.Green()
        b = color.Blue()
        return gc.CreateLinearGradientBrush(x, y, x + w, y + h,
                                            color,
                                            wx.Colour(int(self._gradVal * r + (1.0 - self._gradVal) * 255),
                                                      int(self._gradVal * g + (1.0 - self._gradVal) * 255),
                                                      int(self._gradVal * b + (1.0 - self._gradVal) * 255)))

    def _DrawParent(self, gc, startIndex, endIndex, startIndexRecursive, endIndexRecursive, y, yMax, event, w):
        x0 = startIndexRecursive * w
        x1 = endIndexRecursive * w - 1.0
        y0 = y * (self._eventHeight + self._margin) + self._marginTop
        y1 = y0 + self._eventHeight
        y2 = yMax * (self._eventHeight + self._margin) + self._marginTop - self._margin
        color = self.GetBackgroundColor(event)

        # Overall box
        self._DrawBox(gc,
                      event,
                      x0 - self._margin / 3,
                      y0 - self._margin / 3,
                      x1 + self._margin / 3,
                      y2 + self._margin / 3,
                      wx.Colour(int((color.Red() + self._outlineColorLight[0]) / 2),
                                int((color.Green() + self._outlineColorLight[1]) / 2),
                                int((color.Blue() + self._outlineColorLight[2]) / 2)))

        if startIndex is not None:
            x0 = startIndex * w
        if endIndex is not None:
            x1 = endIndex * w - 1.0

        # Span
        path = gc.CreatePath()
        delta = self._eventHeight / 4
        path.MoveToPoint(x0, y0)
        path.AddLineToPoint(x1, y0)
        path.AddLineToPoint(x1, y1 - delta)
        path.AddLineToPoint(x1 - delta, y1)
        path.AddLineToPoint(x1 - 2 * delta, y1 - delta)
        path.AddLineToPoint(x0 + 2 * delta, y1 - delta)
        path.AddLineToPoint(x0 + delta, y1)
        path.AddLineToPoint(x0, y1 - delta)
        path.CloseSubpath()

        gc.SetBrush(self._Gradient(gc, color, x0, y0, x1 - x0, y1 - y0))
        gc.FillPath(path)

        gc.SetPen(wx.Pen(wx.Colour(*self._outlineColorDark)))
        gc.DrawPath(path)

        x0 = max(0.0, x0)
        x1 = min(self._maxIndex * self._eventWidth, x1)

        # Progress
        x0, y0, x1, y1 = self._DrawProgress(gc, event, x0, y0, x1, y1)

        y1 -= delta

        # Text & icons
        x0, y0, x1, y1 = self._DrawIcons(gc, event, x0, y0, x1, y1)
        self._DrawText(gc, event, x0, y0, x1, y1)

    def _DrawLeaf(self, gc, startIndex, endIndex, yMin, yMax, event, w):
        x0 = startIndex * w
        x1 = endIndex * w - 1.0
        y0 = yMin * (self._eventHeight + self._margin) + self._marginTop
        y1 = yMax * (self._eventHeight + self._margin) + self._marginTop - self._margin

        # Box
        self._DrawBox(gc, event, x0, y0, x1, y1, self.GetBackgroundColor(event))

        x0 = max(0.0, x0)
        x1 = min(self._maxIndex * self._eventWidth, x1)

        # Progress
        x0, y0, x1, y1 = self._DrawProgress(gc, event, x0, y0, x1, y1)

        # Text & icons
        x0, y0, x1, y1 = self._DrawIcons(gc, event, x0, y0, x1, y1)
        self._DrawText(gc, event, x0, y0, x1, y1)

    def _DrawBox(self, gc, event, x0, y0, x1, y1, color):
        outline = wx.Colour(*self._outlineColorLight)

        if event in self._selection:
            outline = wx.BLUE
            color = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)

        path = gc.CreatePath()
        path.AddRoundedRectangle(x0, y0, x1 - x0, y1 - y0, 5.0)
        gc.SetBrush(self._Gradient(gc, color, x0, y0, x1, y1))
        gc.FillPath(path)

        gc.SetPen(wx.Pen(outline))
        gc.DrawPath(path)

    def _DrawProgress(self, gc, event, x0, y0, x1, y1):
        p = self.GetProgress(event)
        if p is not None:
            px0 = x0 + self._eventHeight / 2
            px1 = x1 - self._eventHeight / 2
            py0 = y0 + (self._eventHeight / 4 - self._eventHeight / 8) / 2
            py1 = py0 + self._eventHeight / 8

            gc.SetBrush(wx.Brush(self._outlineColorDark))
            gc.DrawRectangle(px0, py0, px1 - px0, py1 - py0)

            gc.SetBrush(self._Gradient(gc, wx.BLUE, px0, py0, px0 + (px1 - px0) * p, py1))
            gc.DrawRectangle(px0, py0, (px1 - px0) * p, py1 - py0)

            y0 = py1
        return x0, y0, x1, y1

    def _DrawText(self, gc, event, x0, y0, x1, y1):
        gc.SetFont(self.GetFont(event), wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT) if event in self._selection else self.GetForegroundColor(event))
        text = shortenText(gc, self.GetText(event), x1 - x0 - self._margin * 2)
        w, h = gc.GetTextExtent(text)
        gc.DrawText(text, x0 + self._margin, y0 + self._eventHeight / 3 + (y1 - y0 - h - 2 * self._eventHeight / 3) / 2)

    def _DrawIcons(self, gc, event, x0, y0, x1, y1):
        cx = x0
        icons = self.GetIcons(event)
        if icons:
            cx += self._margin
            for icon in icons:
                w = icon.GetWidth()
                h = icon.GetHeight()
                gc.DrawIcon(icon, cx, y0 + (y1 - y0 - h) / 2, w, h)
                cx += w + self._margin
        return cx, y0, x1, y1

    def _GetStartRecursive(self, event):
        dt = self.GetStart(event)
        ls = [] if dt is None else [dt]
        for child in self.GetChildren(event):
            dt = self._GetStartRecursive(child)
            if dt is not None:
                ls.append(dt)
        return min(ls) if ls else None

    def _GetEndRecursive(self, event):
        dt = self.GetEnd(event)
        ls = [] if dt is None else [dt]
        for child in self.GetChildren(event):
            dt = self._GetEndRecursive(child)
            if dt is not None:
                ls.append(dt)
        return max(ls) if ls else None

    def _Invalidate(self):
        self._coords = dict()
        watermark = _Watermark()
        self._maxIndex = int(total_seconds(self._end - self._start) / self._precision / 60)

        def computeEvent(event):
            eventStart = self.GetStart(event)
            eventEnd = self.GetEnd(event)
            eventRStart = self._GetStartRecursive(event)
            eventREnd = self._GetEndRecursive(event)

            if eventRStart is not None and eventREnd is not None and not (eventRStart >= self._end or eventREnd < self._start):
                rstart = int(math.floor(total_seconds(eventRStart - self._start) / self._precision / 60))
                start = None if eventStart is None else int(math.floor(total_seconds(eventStart - self._start) / self._precision / 60))
                rend = int(math.floor(total_seconds(eventREnd - self._start) / self._precision / 60))
                end = None if eventEnd is None else int(math.floor(total_seconds(eventEnd - self._start) / self._precision / 60))
                if rend > rstart:
                    y = watermark.height(rstart, rend)
                    watermark.add(rstart, rend, y + 1)
                    yMax = y + 1
                    for child in self.GetChildren(event):
                        yMax = max(yMax, computeEvent(child))
                    self._coords[event] = (start, end, rstart, rend, y, yMax)
                    return yMax

        for rootEvent in self.GetRootEvents():
            computeEvent(rootEvent)

        bmp = wx.EmptyBitmap(10, 10) # Don't care
        memDC = wx.MemoryDC()
        memDC.SelectObject(bmp)
        try:
            gc = wx.GraphicsContext.Create(memDC)
            gc.SetFont(wx.NORMAL_FONT, wx.BLACK)

            self._headerSpans = []
            self._daySpans = []
            startIdxHeader = 0
            startIdxDay = 0
            currentFmt = self.FormatDateTime(self._start)
            currentDay = self._start.date()
            headerWidth = gc.GetTextExtent(currentFmt)[0]
            for idx in xrange(1, self._maxIndex):
                dateTime = self._start + datetime.timedelta(minutes=self._precision * idx)
                fmt = self.FormatDateTime(dateTime)
                if fmt != currentFmt:
                    headerWidth += gc.GetTextExtent(fmt)[0]
                    self._headerSpans.append((startIdxHeader, idx))
                    startIdxHeader = idx
                    currentFmt = fmt
                if dateTime.date() != currentDay:
                    self._daySpans.append((startIdxDay, idx))
                    startIdxDay = idx
                    currentDay = dateTime.date()
            self._headerSpans.append((startIdxHeader, self._maxIndex))
            self._daySpans.append((startIdxDay, self._maxIndex))
            headerWidth += self._margin * 2 * len(self._headerSpans)

            self._minSize = (int(max(headerWidth, self._eventWidthMin * self._maxIndex)),
                             self._marginTop + (watermark.totalHeight() - 1) * (self._eventHeight + self._margin))
            self._OnResize()
        finally:
            memDC.SelectObject(wx.NullBitmap)


class CalendarPrintout(wx.Printout):
    def __init__(self, calendar, settings, *args, **kwargs):
        super(CalendarPrintout, self).__init__(*args, **kwargs)
        self._calendar = calendar
        self._settings = settings
        self._count = None

    def _PageCount(self):
        if self._count is None:
            minW, minH = self._calendar._minSize
            dc = self.GetDC()
            dcw, dch = dc.GetSizeTuple()
            cw = minW
            ch = minW * dch / dcw
            cells = int(math.ceil(1.0 * (ch - self._calendar._marginTop) / (self._calendar._eventHeight + self._calendar._margin)))
            total = int(math.ceil(1.0 * (minH - self._calendar._marginTop) / (self._calendar._eventHeight + self._calendar._margin))) + 1
            self._count = int(math.ceil(1.0 * total / cells))
        return self._count

    def GetPageInfo(self):
        return 1, self._PageCount(), 1, 1

    def HasPage(self, page):
        return page <= self._PageCount()

    def OnPrintPage(self, page):
        # Cannot print with a GraphicsContext...
        minW, minH = self._calendar._minSize
        dc = self.GetDC()
        dcw, dch = dc.GetSizeTuple()
        cw = minW
        ch = minW * dch / dcw
        cells = int(math.ceil(1.0 * (ch - self._calendar._marginTop) / (self._calendar._eventHeight + self._calendar._margin)))
        dy = 1.0 * cells * (self._calendar._eventHeight + self._calendar._margin) * (page - 1)

        bmp = wx.EmptyBitmap(cw, ch)
        memDC = wx.MemoryDC()
        memDC.SelectObject(bmp)
        try:
            memDC.SetBackground(wx.WHITE_BRUSH)
            memDC.Clear()

            oldWidth = self._calendar._eventWidth
            self._calendar._eventWidth = self._calendar._eventWidthMin
            try:
                gc = wx.GraphicsContext.Create(memDC)
                self._calendar._Draw(gc, cw, ch, 0, dy)
            finally:
                self._calendar._eventWidth = oldWidth
            dc.SetUserScale(1.0 * dcw / cw, 1.0 * dch / ch)
            dc.Blit(0, 0, cw, ch, memDC, 0, 0)
        finally:
            memDC.SelectObject(wx.NullBitmap)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.domain import base
from taskcoachlib import patterns


class CategoryList(base.Collection):
    @patterns.eventSource    
    def extend(self, categories, event=None):
        super(CategoryList, self).extend(categories, event=event)
        for category in self._compositesAndAllChildren(categories):
            for categorizable in category.categorizables():
                categorizable.addCategory(category, event=event, modify=False)

    @patterns.eventSource                
    def removeItems(self, categories, event=None):
        super(CategoryList, self).removeItems(categories, event=event)
        for category in self._compositesAndAllChildren(categories):
            for categorizable in category.categorizables():
                categorizable.removeCategory(category, event=event)

    def findCategoryByName(self, name):
        for category in self:
            recursive = ' -> ' in name
            if category.subject(recursive=recursive) == name:
                return category
        return None
    
    def filteredCategories(self):
        return [category for category in self if category.isFiltered()]
    
    @patterns.eventSource
    def resetAllFilteredCategories(self, event=None):
        for category in self:
            category.setFiltered(False, event=event)



"""
test_core.py

Created by Kang Zhang on 2009-08-09
"""

from __future__ import with_statement

import os
import tempfile
import shutil

from keyring.tests.py30compat import unittest

import mock

import keyring.backend
import keyring.core
from keyring import errors

PASSWORD_TEXT = "This is password"
PASSWORD_TEXT_2 = "This is password2"
KEYRINGRC = "keyringrc.cfg"


class TestKeyring(keyring.backend.KeyringBackend):
    """A faked keyring for test.
    """
    def __init__(self):
        self.passwords = {}

    def supported(self):
        return 0

    def get_password(self, service, username):
        return PASSWORD_TEXT

    def set_password(self, service, username, password):
        self.passwords[(service, username)] = password
        return 0

    def delete_password(self, service, username):
        try:
            del self.passwords[(service, username)]
        except KeyError:
            raise errors.PasswordDeleteError("not set")


class TestKeyring2(TestKeyring):
    """Another faked keyring for test.
    """
    def get_password(self, service, username):
        return PASSWORD_TEXT_2


class CoreTestCase(unittest.TestCase):
    mock_global_backend = mock.patch('keyring.core._keyring_backend')

    @mock_global_backend
    def test_set_password(self, backend):
        """
        set_password on the default keyring is called.
        """
        keyring.core.set_password("test", "user", "passtest")
        backend.set_password.assert_called_once_with('test', 'user',
            'passtest')

    @mock_global_backend
    def test_get_password(self, backend):
        """
        set_password on the default keyring is called.
        """
        result = keyring.core.get_password("test", "user")
        backend.get_password.assert_called_once_with('test', 'user')
        self.assertIsNotNone(result)

    @mock_global_backend
    def test_delete_password(self, backend):
        keyring.core.delete_password("test", "user")
        backend.delete_password.assert_called_once_with('test', 'user')

    def test_set_keyring_in_runtime(self):
        """Test the function of set keyring in runtime.
        """
        keyring.core.set_keyring(TestKeyring())

        keyring.core.set_password("test", "user", "password")
        self.assertEqual(keyring.core.get_password("test", "user"),
            PASSWORD_TEXT)

    def test_set_keyring_in_config(self):
        """Test setting the keyring by config file.
        """
        # create the config file
        config_file = open(KEYRINGRC, 'w')
        config_file.writelines([
            "[backend]\n",
            # the path for the user created keyring
            "keyring-path= %s\n" % os.path.dirname(os.path.abspath(__file__)),
            # the name of the keyring class
            "default-keyring=test_core.TestKeyring2\n",
            ])
        config_file.close()

        # init the keyring lib, the lib will automaticlly load the
        # config file and load the user defined module
        keyring.core.init_backend()

        keyring.core.set_password("test", "user", "password")
        self.assertEqual(keyring.core.get_password("test", "user"),
            PASSWORD_TEXT_2)

        os.remove(KEYRINGRC)

    def test_load_config(self):
        tempdir = tempfile.mkdtemp()
        old_location = os.getcwd()
        os.chdir(tempdir)
        personal_cfg = os.path.join(os.path.expanduser("~"), "keyringrc.cfg")
        if os.path.exists(personal_cfg):
            os.rename(personal_cfg, personal_cfg + '.old')
            personal_renamed = True
        else:
            personal_renamed = False

        # loading with an empty environment
        keyring.core.load_config()

        # loading with a file that doesn't have a backend section
        cfg = os.path.join(tempdir, "keyringrc.cfg")
        f = open(cfg, 'w')
        f.write('[keyring]')
        f.close()
        keyring.core.load_config()

        # loading with a file that doesn't have a default-keyring value
        cfg = os.path.join(tempdir, "keyringrc.cfg")
        f = open(cfg, 'w')
        f.write('[backend]')
        f.close()
        keyring.core.load_config()

        os.chdir(old_location)
        shutil.rmtree(tempdir)
        if personal_renamed:
            os.rename(personal_cfg + '.old', personal_cfg)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(CoreTestCase))
    return suite

if __name__ == "__main__":
    unittest.main(defaultTest="test_suite")



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib.domain import task, date
from taskcoachlib import config


class TaskStatusTest(test.TestCase):    
    def setUp(self):
        self.settings = task.Task.settings = config.Settings(load=False)
        self.now = date.Now()
        self.yesterday = self.now - date.ONE_DAY
        self.nearFuture = self.now + date.ONE_DAY - date.ONE_HOUR
        self.dates = (self.yesterday, self.nearFuture)
        self.farFuture = self.now + date.ONE_DAY + date.ONE_DAY

    def assertTaskStatus(self, status, **taskKwArgs):
        self.assertEqual(status, task.Task(**taskKwArgs).status())
        
    # No dates/times
    
    def testDefaultTaskIsInactive(self):
        self.assertTaskStatus(task.status.inactive)
    
    # One date/time
        
    def testTaskWithCompletionInThePastIsCompleted(self):
        self.assertTaskStatus(task.status.completed, completionDateTime=self.yesterday)
        
    def testTaskWithCompletionInTheFutureIsCompleted(self):
        # Maybe keep the task inactive until the completion date passes? 
        # That would be more consistent with the other date/times
        self.assertTaskStatus(task.status.completed, completionDateTime=self.nearFuture)

    def testTaskWithPlannedStartInThePastIsLate(self):
        self.assertTaskStatus(task.status.late, plannedStartDateTime=self.yesterday)
                
    def testTaskWithPlannedStartInTheFutureIsInactive(self):
        self.assertTaskStatus(task.status.inactive, plannedStartDateTime=self.nearFuture)
        
    def testTaskWithActualStartInThePastIsActive(self):
        self.assertTaskStatus(task.status.active, actualStartDateTime=self.yesterday)

    def testTaskWithActualStartInTheFutureIsInactive(self):
        self.assertTaskStatus(task.status.inactive, actualStartDateTime=self.nearFuture)
        
    def testTaskWithDueInThePastIsOverdue(self):
        self.assertTaskStatus(task.status.overdue, dueDateTime=self.yesterday)

    def testTaskWithDueInTheFutureIsInactive(self):
        self.assertTaskStatus(task.status.inactive, dueDateTime=self.farFuture)
        
    def testTaskWithDueInTheNearFutureIsDueSoon(self):
        self.assertTaskStatus(task.status.duesoon, dueDateTime=self.nearFuture)
    
    # Two dates/times
    
    # planned start date/time and actual start date/time
        
    def testTaskWithPlannedAndActualStartInThePastIsActive(self):
        self.assertTaskStatus(task.status.active, 
                              plannedStartDateTime=self.yesterday,
                              actualStartDateTime=self.yesterday)
        
    def testTaskWithPlannedStartInThePastAndActualStartInTheFutureIsLate(self):
        self.assertTaskStatus(task.status.late, 
                              plannedStartDateTime=self.yesterday,
                              actualStartDateTime=self.nearFuture)
        
    def testTaskWithPlannedStartInTheFutureAndActualStartInThePastIsActive(self):
        self.assertTaskStatus(task.status.active, 
                              plannedStartDateTime=self.nearFuture,
                              actualStartDateTime=self.yesterday)

    def testTaskWithPlannedAndActualStartInTheFutureIsInactive(self):
        self.assertTaskStatus(task.status.inactive, 
                              plannedStartDateTime=self.nearFuture,
                              actualStartDateTime=self.nearFuture)
    
    # planned start date/time and due date/time
        
    def testTaskWithPlannedStartAndDueInThePastIsOverdue(self):
        self.assertTaskStatus(task.status.overdue, 
                              plannedStartDateTime=self.yesterday,
                              dueDateTime=self.yesterday)

    def testTaskWithPlannedStartInThePastAndDueInTheFutureIsLate(self):
        self.assertTaskStatus(task.status.late, 
                              plannedStartDateTime=self.yesterday,
                              dueDateTime=self.farFuture)
       
    def testTaskWithPlannedStartInThePastAndDueInTheNearFutureIsDueSoon(self):
        self.assertTaskStatus(task.status.duesoon, 
                              plannedStartDateTime=self.yesterday,
                              dueDateTime=self.nearFuture)
       
    def testTaskWithPlannedStartInTheFutureAndDueInThePastIsOverdue(self):
        self.assertTaskStatus(task.status.overdue, 
                              plannedStartDateTime=self.nearFuture,
                              dueDateTime=self.yesterday)

    def testTaskWithPlannedStartInTheFutureAndDueInTheFutureIsLate(self):
        self.assertTaskStatus(task.status.inactive, 
                              plannedStartDateTime=self.nearFuture,
                              dueDateTime=self.farFuture)
       
    def testTaskWithPlannedStartInTheFutureAndDueInTheNearFutureIsDueSoon(self):
        self.assertTaskStatus(task.status.duesoon, 
                              plannedStartDateTime=self.nearFuture,
                              dueDateTime=self.nearFuture)

    # planned start date/time and completion date/time
    
    def testTaskWithPlannedStartAndCompletionInThePastIsCompleted(self):
        self.assertTaskStatus(task.status.completed, 
                              plannedStartDateTime=self.yesterday,
                              completionDateTime=self.yesterday)

    def testTaskWithPlannedStartInThePastAndCompletionInTheFutureIsCompleted(self):
        self.assertTaskStatus(task.status.completed, 
                              plannedStartDateTime=self.yesterday,
                              completionDateTime=self.nearFuture)

    def testTaskWithPlannedStartInTheFutureAndCompletionInThePastIsCompleted(self):
        self.assertTaskStatus(task.status.completed, 
                              plannedStartDateTime=self.nearFuture,
                              completionDateTime=self.yesterday)

    def testTaskWithPlannedStartInTheFutureAndCompletionInTheFutureIsComplete(self):
        self.assertTaskStatus(task.status.completed, 
                              plannedStartDateTime=self.nearFuture,
                              completionDateTime=self.nearFuture)
    
    # actual start date/time and due date/time
    
    def testTaskWithActualStartAndDueInThePastIsOverdue(self):
        self.assertTaskStatus(task.status.overdue, 
                              actualStartDateTime=self.yesterday,
                              dueDateTime=self.yesterday)

    def testTaskWithActualStartInThePastAndDueInTheFutureIsActive(self):
        self.assertTaskStatus(task.status.active, 
                              actualStartDateTime=self.yesterday,
                              dueDateTime=self.farFuture)

    def testTaskWithActualStartInThePastAndDueInTheNearFutureIsDueSoon(self):
        self.assertTaskStatus(task.status.duesoon, 
                              actualStartDateTime=self.yesterday,
                              dueDateTime=self.nearFuture)

    def testTaskWithActualStartInTheFutureAndDueInThePastIsOverdue(self):
        self.assertTaskStatus(task.status.overdue, 
                              actualStartDateTime=self.nearFuture,
                              dueDateTime=self.yesterday)

    def testTaskWithActualStartInTheFutureAndDueInTheFutureIsActive(self):
        self.assertTaskStatus(task.status.inactive, 
                              actualStartDateTime=self.nearFuture,
                              dueDateTime=self.farFuture)

    def testTaskWithActualStartInTheFutureAndDueInTheNearFutureIsDueSoon(self):
        self.assertTaskStatus(task.status.duesoon, 
                              actualStartDateTime=self.nearFuture,
                              dueDateTime=self.nearFuture)

    # actual start date/time and completion date/time
   
    def testTaskWithActualStartAndCompletionInThePastIsCompleted(self):
        self.assertTaskStatus(task.status.completed, 
                              actualStartDateTime=self.yesterday,
                              completionDateTime=self.yesterday)

    def testTaskWithActualStartInThePastAndCompletionInTheFutureIsCompleted(self):
        self.assertTaskStatus(task.status.completed, 
                              actualStartDateTime=self.yesterday,
                              completionDateTime=self.nearFuture)

    def testTaskWithActualStartInTheFutureAndCompletionInThePastIsCompleted(self):
        self.assertTaskStatus(task.status.completed, 
                              actualStartDateTime=self.nearFuture,
                              completionDateTime=self.yesterday)

    def testTaskWithActualStartInTheFutureAndCompletionInTheFutureIsComplete(self):
        self.assertTaskStatus(task.status.completed, 
                              actualStartDateTime=self.nearFuture,
                              completionDateTime=self.nearFuture)
   
    # due date/time and completion date/time
    
    def testTaskWithDueAndCompletionInThePastIsCompleted(self):
        self.assertTaskStatus(task.status.completed, 
                              dueDateTime=self.yesterday,
                              completionDateTime=self.yesterday)

    def testTaskWithDueInThePastAndCompletionInTheFutureIsCompleted(self):
        self.assertTaskStatus(task.status.completed, 
                              dueDateTime=self.yesterday,
                              completionDateTime=self.nearFuture)

    def testTaskWithDueInTheFutureAndCompletionInThePastIsCompleted(self):
        self.assertTaskStatus(task.status.completed, 
                              dueDateTime=self.nearFuture,
                              completionDateTime=self.yesterday)

    def testTaskWithDueInTheFutureAndCompletionInTheFutureIsComplete(self):
        self.assertTaskStatus(task.status.completed, 
                              dueDateTime=self.nearFuture,
                              completionDateTime=self.nearFuture)
   
    # Three dates/times
    
    # planned start date/time, actual start date/time and due date/time
    # (Other combinations are not interesting since they are always completed)
    
    def testTaskIsOverdueWheneverDueIsInThePast(self):
        for planned in self.dates:
            for actual in self.dates:
                self.assertTaskStatus(task.status.overdue, 
                                      plannedStartDateTime=planned,
                                      actualStartDateTime=actual,
                                      dueDateTime=self.yesterday)

    def testTaskIsDuesoonWheneverDueIsInTheNearFuture(self):
        for planned in self.dates:
            for actual in self.dates:
                self.assertTaskStatus(task.status.duesoon, 
                                      plannedStartDateTime=planned,
                                      actualStartDateTime=actual,
                                      dueDateTime=self.nearFuture)
         
    def testTaskIsOverdueWheneverDueIsInTheFuture(self):
        for planned in self.dates:
            expectedStatusBasedOnPlannedStart = task.status.late if planned < self.now else task.status.inactive
            for actual in self.dates:
                expectedStatus = task.status.active if actual < self.now else expectedStatusBasedOnPlannedStart
                self.assertTaskStatus(expectedStatus, 
                                      plannedStartDateTime=planned,
                                      actualStartDateTime=actual,
                                      dueDateTime=self.farFuture)
               
    # Four date/times (always completed)
    
    def testTaskWithCompletionDateTimeIsAlwaysCompleted(self):
        for planned in self.dates:
            for actual in self.dates:
                for due in self.dates + (self.farFuture,):
                    for completion in self.dates:
                        self.assertTaskStatus(task.status.completed, 
                                              plannedStartDateTime=planned,
                                              actualStartDateTime=actual,
                                              dueDateTime=due,
                                              completionDateTime=completion)

    # Prerequisites
    
    def testTaskWithUncompletedPrerequisiteIsNeverLate(self):
        prerequisite = task.Task()
        for planned in self.dates:
            self.assertTaskStatus(task.status.inactive, 
                                  plannedStartDateTime=planned,
                                  prerequisites=[prerequisite])

    def testTaskWithCompletedPrerequisiteIsLateWhenPlannedStartIsInThePast(self):
        prerequisite = task.Task(completionDateTime=self.yesterday)
        for planned in self.dates:
            expectedStatus = task.status.late if planned < self.now else task.status.inactive
            self.assertTaskStatus(expectedStatus, plannedStartDateTime=planned,
                                  prerequisites=[prerequisite])
            
    def testMutualPrerequisites(self):
        taskA = task.Task()
        taskB = task.Task(prerequisites=[taskA])
        taskA.addPrerequisites([taskB])
        for eachTask in (taskA, taskB):
            self.assertEqual(task.status.inactive, eachTask.status())



'''
Task Coach - Your friendly task manager
Copyright (C) 2012 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
from taskcoachlib.i18n import _
from taskcoachlib.widgets import dialog


class XFCE4WarningDialog(dialog.Dialog):
    def __init__(self, parent, settings):
        self.__settings = settings
        super(XFCE4WarningDialog, self).__init__(parent, _('Warning'),
                                                 buttonTypes=wx.OK)

    def createInterior(self):
        return wx.Panel(self._panel)

    def fillInterior(self):
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(wx.StaticText(self._interior, label=_('Task Coach has known issues with XFCE4 session management.\n') + \
                                _('If you experience random freeze at startup, please uncheck\nthe "Use X11 session management" in the Features tab of the preferences.\n')))
        self._checkbox = wx.CheckBox(self._interior, label=_('Do not show this dialog at startup')) # pylint: disable=W0201
        self._checkbox.SetValue(True)
        sizer.Add(self._checkbox)
        self._interior.SetSizer(sizer)

    def ok(self, event=None):
        self.__settings.setboolean('feature', 'showsmwarning', not self._checkbox.GetValue())
        super(XFCE4WarningDialog, self).ok(event)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import os
import struct
import sys
import tempfile
import wx

import taskcoachlib.thirdparty.gntp.notifier as Growl
from taskcoachlib import meta
from notifier import AbstractNotifier


class GrowlNotifier(AbstractNotifier):
    def __init__(self):
        super(GrowlNotifier, self).__init__()
        try:
            # pylint: disable=E1101
            self._notifier = Growl.GrowlNotifier(applicationName=meta.name, notifications=[u'Reminder'])
            self._notifier.register()
        except:
            self._available = False  # pylint: disable=W0702
        else:
            self._available = True

    def getName(self):
        return 'Growl'

    def isAvailable(self):
        return self._available

    def notify(self, title, summary, bitmap, **kwargs):
        # Not really efficient...
        fd, filename = tempfile.mkstemp('.png')
        os.close(fd)
        try:
            bitmap.SaveFile(filename, wx.BITMAP_TYPE_PNG)
            self._notifier.notify(noteType=u'Reminder', icon=file(filename, 'rb').read(), title=title, description=summary,
                                  sticky=True)
        finally:
            os.remove(filename)


AbstractNotifier.register(GrowlNotifier())



'''
Task Coach - Your friendly task manager
Copyright (C) 2013 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import weakref
import test, mock, os


class LeakTest(test.TestCase):
    def setUp(self):
        self.mockApp = mock.App()
        self.mockApp.addTask()

    def tearDown(self):
        self.mockApp.iocontroller.saveas('Test.tsk')
        os.remove('Test.tsk')
        self.mockApp.quitApplication()
        mock.App.deleteInstance()
        super(LeakTest, self).tearDown()

    def testClear(self):
        """taskRef = weakref.ref(self.mockApp.task)
        del self.mockApp.task
        self.mockApp.taskFile.clear()
        self.failUnless(taskRef() is None)"""



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.persistence.icalendar import ical
from taskcoachlib.domain import task
from taskcoachlib import meta


def extendedWithAncestors(selection):
    extendedSelection = selection[:]
    for item in selection:
        for ancestor in item.ancestors():
            if ancestor not in extendedSelection:
                extendedSelection.append(ancestor)
    return extendedSelection


class iCalendarWriter(object):
    def __init__(self, fd, filename=None):
        self.__fd = fd

    def write(self, viewer, settings, selectionOnly=False): # pylint: disable=W0613
        items = viewer.visibleItems()
        if selectionOnly:
            selection = viewer.curselection()
            if viewer.isTreeViewer():
                selection = extendedWithAncestors(selection)
            items = [item for item in items if item in selection]
        return self.writeItems(items)
    
    def writeItems(self, items):
        self.__fd.write('BEGIN:VCALENDAR\r\n')
        self._writeMetaData()
        count = 0
        for item in items:
            transform = ical.VCalFromTask if isinstance(item, task.Task) else ical.VCalFromEffort
            self.__fd.write(transform(item, encoding=False))
            count += 1
        self.__fd.write('END:VCALENDAR\r\n')
        return count

    def _writeMetaData(self):
        self.__fd.write('VERSION:2.0\r\n')
        domain = meta.url[len('http://'):].strip('/')
        self.__fd.write('PRODID:-//%s//NONSGML %s V%s//EN\r\n'%(domain, 
                                                                meta.name, 
                                                                meta.version))
